import os
import sys
import json
import logging
import re
import numpy as np
import pandas as pd
import glob
from scipy.io import mmwrite
from scipy.sparse import csr_matrix
from celescope.tools.report import reporter
from celescope.tools.utils import *
from celescope.tools.Analysis import Analysis
from .otsu import *
import celescope.tools

toolsdir = os.path.dirname(celescope.tools.__file__)

class Analysis_capture_virus(Analysis):

    def get_virus_tsne(self):
        df = pd.merge(self.tsne_df, self.virus_df, on="barcode", how="left")
        self.df_file = f'{self.outdir}/{self.sample}_tsne.tsv'
        df.to_csv(self.df_file, sep='\t')
        df["UMI"] = df["UMI"].fillna(0)
        tSNE_1 = list(df.tSNE_1)
        tSNE_2 = list(df.tSNE_2)
        virus_UMI = list(df.UMI)
        res = {"tSNE_1": tSNE_1, "tSNE_2": tSNE_2, "virus_UMI": virus_UMI}
        return res

    def get_virus_df(self):
        self.virus_df = pd.read_csv(self.virus_file, sep="\t")

    def report(self):
        t = reporter(
            name=self.step,
            assay=self.assay,
            sample=self.sample,
            outdir=self.outdir + '/..',
        )
        t.get_report()

    def run(self):
        cluster_tsne = self.get_cluster_tsne(colname='cluster')
        self.get_virus_df()
        virus_tsne = self.get_virus_tsne()
        table_dict = self.get_marker_gene_table()
        self.report_prepare(
            cluster_tsne=cluster_tsne,
            virus_tsne=virus_tsne,
            table_dict=table_dict,
        )
        self.report()

@add_log
def otsu_thresh(virus_file, outdir, sample):
    df = pd.read_csv(virus_file, sep='\t')
    array = np.log10(df["UMI"])
    hist = array2hist(array)
    thresh = threshold_otsu(hist)
    fname = f'{outdir}/{sample}_otsu_plot.png'
    makePlot(hist, thresh, fname)

    threshold = int(10 ** thresh)
    otsu_thresh.logger.info(f'otsu UMI threshold: {threshold}')
    df_thresh = df[df["UMI"] >= threshold]
    otsu_virus_file = f'{outdir}/{sample}_otsu_count.tsv'
    df_thresh.to_csv(otsu_virus_file, sep='\t')
    return otsu_virus_file


@add_log
def analysis_capture_virus(args):

    # check dir
    outdir = args.outdir
    sample = args.sample
    virus_file = args.virus_file
    match_dir = args.match_dir
    assay = args.assay
    umi_threshold = args.umi_threshold

    if not os.path.exists(outdir):
        os.system('mkdir -p %s' % (outdir))

    if umi_threshold == 'otsu':
        virus_file = otsu_thresh(virus_file, outdir, sample)        

    ana = Analysis_capture_virus(     
        sample,
        outdir,
        assay,
        match_dir=match_dir,
        step='analysis_capture_virus',     
    )
    ana.virus_file = virus_file
    ana.run()


def get_opts_analysis_capture_virus(parser, sub_program):
    parser.add_argument("--umi_threshold", help='method to find virus UMI threshold', 
        choices=['otsu', 'none'], default='otsu')
    if sub_program:
        s_common(parser)
        parser.add_argument('--match_dir', help='match_dir', required=True)
        parser.add_argument('--virus_file', help='virus UMI count file', required=True)
        



import unittest
import pandas as pd
import numpy as np
import os
from .otsu import *

class test_capture(unittest.TestCase):
    def setUp(self):
        pass

    def test_otsu(self):
        count_files = [
'/SGRNJ02/RandD4/virus_panel/20210124_4/S1225_EBV_Skin_Auto_SDF_NEB/07.count_virus/S1225_EBV_Skin_Auto_SDF_NEB_virus_UMI_count.tsv',
'/SGRNJ02/RandD4/virus_panel/20210124/virus_test3_R_A_Beads_Manual_KZ/04.count_capture_virus/virus_test3_R_A_Beads_Manual_KZ_virus_UMI_count.tsv',
'/SGRNJ02/RandD4/virus_panel/20210124/virus_test3_R_A_3Mins_Manual_KZ/04.count_capture_virus/virus_test3_R_A_3Mins_Manual_KZ_virus_UMI_count.tsv'
        ]
        count_file = '/SGRNJ02/RandD4/virus_panel/20210124_4/S1225_EBV_Skin_Auto_SDF_NEB/07.count_virus/S1225_EBV_Skin_Auto_SDF_NEB_virus_UMI_count.tsv'
        for count_file in count_files:
            df = pd.read_csv(count_file, sep='\t')
            array = np.log10(df["UMI"])
            hist = array2hist(array)
            print(hist)
            thresh = threshold_otsu(hist)
            fname = os.path.basename(count_file) + '.png'
            makePlot(hist, thresh, fname)



__STEPS__ = [    
    'sample',
    'barcode',
    'cutadapt',
    "STAR_virus",
    "count_capture_virus",
    "analysis_capture_virus",
]
__ASSAY__ = 'capture_virus'

IMPORT_DICT = {
    'STAR_virus': 'celescope.rna_virus',
}


from collections import defaultdict
import os
import pysam
from .utils import *

class Chemistry():

    def __init__(self, fq1):
        self.fq1 = fq1
        self.fq1_list = fq1.split(',')
        self.nRead = 10000

    @add_log
    def check_chemistry(self):
        chemistry_list = []
        for fastq1 in self.fq1_list:
            print(fastq1)
            chemistry = self.get_chemistry(fastq1)
            chemistry_list.append(chemistry)
        if len(set(chemistry_list)) != 1:
           Chemistry.check_chemistry.logger.warning('multiple chemistry found!' + str(chemistry_list))
        return chemistry_list

    @add_log
    def get_chemistry(self, fq1):
        '''
        'scopeV2.0.1': 'C8L16C8L16C8L1U8T18'
        'scopeV2.1.1': 'C8L16C8L16C8L1U12T18'
        'scopeV2.2.1': 'C8L16C8L16C8L1U12T18' with 4 types of linkers
        '''
        # init
        linker_4_file, _whitelist = get_scope_bc('scopeV2.2.1')
        linker_4_list, _num = read_one_col(linker_4_file)
        linker_4_dict = defaultdict(int)
        linker_wrong_dict = defaultdict(int)
        pattern_dict = parse_pattern('C8L16C8L16C8L1U12T18')
        T4_n = 0
        L57C_n = 0

        with pysam.FastxFile(fq1) as fh:
            for i in range(self.nRead):
                entry = fh.__next__()
                seq = entry.sequence
                L57C = seq[56]
                if L57C == 'C':
                    L57C_n += 1
                T4 = seq[65:69]
                if T4 == 'TTTT':
                    T4_n += 1
                linker = seq_ranges(seq, pattern_dict=pattern_dict['L'])
                if linker in linker_4_list:
                    linker_4_dict[linker] += 1
                else:
                    linker_wrong_dict[linker] += 1

        percent_T4 =  T4_n / self.nRead 
        percent_L57C = L57C_n / self.nRead
        Chemistry.get_chemistry.logger.info(f'percent T4: {percent_T4}')
        Chemistry.get_chemistry.logger.info(f'percent L57C: {percent_L57C}')
        if percent_T4 > 0.5:
            chemistry = 'scopeV2.0.1'
        else:
            # V2.1.1 or V2.2.1 or failed
            valid_linker_type = 0
            for linker in linker_4_dict:
                linker_4_dict[linker] = linker_4_dict[linker] / self.nRead
                if linker_4_dict[linker] > 0.05:
                    valid_linker_type += 1
            Chemistry.get_chemistry.logger.info(linker_4_dict)
            if valid_linker_type == 0:
                print(linker_wrong_dict)
                raise Exception('auto chemistry detection failed!')
            elif valid_linker_type == 1:
                chemistry = 'scopeV2.1.1'
            elif valid_linker_type < 4:
                chemistry = 'scopeV2.2.1'
                Chemistry.get_chemistry.logger.warning(
                    f'chemistry scopeV2.2.1 only has {valid_linker_type} linker types!')
            else:
                chemistry = 'scopeV2.2.1'
        Chemistry.get_chemistry.logger.info(f'chemistry: {chemistry}')
        return chemistry






import logging
import os
import glob
import sys
import re
import subprocess
import importlib
import resource
import time
import argparse
import gzip
import itertools
from datetime import timedelta
from collections import defaultdict
from functools import wraps
from collections import Counter
import json

import pandas as pd
import numpy as np
import pysam
import xopen
import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt

import celescope.tools
from celescope.tools.__init__ import __PATTERN_DICT__

tools_dir = os.path.dirname(celescope.tools.__file__)


def add_log(func):
    '''
    logging start and done.
    '''
    logging.basicConfig(
        level=logging.INFO,
        stream=sys.stdout,
        format='%(asctime)s - %(name)s - %(levelname)s - %(message)s'
    )
    module = func.__module__
    name = func.__name__
    logger_name = f'{module}.{name}'
    logger = logging.getLogger(logger_name)

    @wraps(func)
    def wrapper(*args, **kwargs):
        if args and hasattr(args[0], 'debug') and args[0].debug:
            logger.setLevel(10) # debug

        logger.info('start...')
        start = time.time()
        result = func(*args, **kwargs)
        end = time.time()
        used = timedelta(seconds=end - start)
        logger.info(f'done. time used: {used}')
        return result

    wrapper.logger = logger
    return wrapper


def using(point=""):
    usage=resource.getrusage(resource.RUSAGE_SELF)
    return '''%s: usertime=%s systime=%s mem=%s mb
        '''%(point,usage[0],usage[1],
                usage[2]/1024.0)


def add_mem(func):
    '''
    logging mem.
    '''
    logging.basicConfig(
        level=logging.INFO,
        format='%(asctime)s - %(name)s - %(levelname)s - %(message)s'
    )
    module = func.__module__
    name = func.__name__
    logger_name = f'{module}.{name}'
    logger = logging.getLogger(logger_name)

    @wraps(func)
    def wrapper(*args, **kwargs):
        logger.info(using("before"))
        result = func(*args, **kwargs)
        logger.info(using("after"))
        return result

    wrapper.logger = logger
    return wrapper


def arg_str(arg, arg_name):
    '''
    return action store_true arguments as string
    '''
    if arg:
        return '--' + arg_name
    return ''


def read_barcode_file(match_dir, return_file=False):
    '''
    multi version compatible
    '''
    match_barcode_file1 = glob.glob(
        f"{match_dir}/*count*/*_cellbarcode.tsv")
    match_barcode_file2 = glob.glob(
        f"{match_dir}/*count*/*matrix_10X/*_cellbarcode.tsv")
    match_barcode_file3 = glob.glob(
        f"{match_dir}/*count*/*matrix_10X/*barcodes.tsv")
    match_barcode_file = (
        match_barcode_file1 +
        match_barcode_file2 +
        match_barcode_file3)[0]
    match_barcode, cell_total = read_one_col(match_barcode_file)
    match_barcode = set(match_barcode)
    if return_file:
        return match_barcode, cell_total, match_barcode_file
    return match_barcode, cell_total


def format_stat(count, total_count):
    percent = round(count / total_count * 100, 2)
    string = f'{format_number(count)}({percent}%)'
    return string


def multi_opts(assay):
    readme = f'{assay} multi-samples'
    parser = argparse.ArgumentParser(readme)
    parser.add_argument('--mod', help='mod, sjm or shell', choices=['sjm', 'shell'], default='sjm')
    parser.add_argument(
        '--mapfile',
        help='''
            tsv file, 4 columns:
            1st col: LibName;
            2nd col: DataDir;
            3rd col: SampleName;
            4th col: Cell number or match_dir, optional;
        ''',
        required=True)
    parser.add_argument('--chemistry', choices=__PATTERN_DICT__.keys(), help='chemistry version', default='auto')
    parser.add_argument('--whitelist', help='cellbarcode list')
    parser.add_argument('--linker', help='linker')
    parser.add_argument('--pattern', help='read1 pattern')
    parser.add_argument('--outdir', help='output dir', default="./")
    parser.add_argument(
        '--adapt',
        action='append',
        help='adapter sequence',
        default=[
            'polyT=A{15}',
            'p5=AGATCGGAAGAGCACACGTCTGAACTCCAGTCAC'])
    parser.add_argument(
        '--minimum_length',
        dest='minimum_length',
        help='minimum_length',
        default=20)
    parser.add_argument(
        '--nextseq-trim',
        dest='nextseq_trim',
        help='nextseq_trim',
        default=20)
    parser.add_argument(
        '--overlap',
        help='minimum overlap length, default=5',
        default=5)
    parser.add_argument(
        '--lowQual',
        type=int,
        help='max phred of base as lowQual',
        default=0)
    parser.add_argument(
        '--lowNum',
        type=int,
        help='max number with lowQual allowed',
        default=2)
    parser.add_argument(
        '--rm_files',
        action='store_true',
        help='remove redundant fq.gz and bam after running')
    parser.add_argument('--steps_run', help='steps to run', default='all')
    parser.add_argument('--debug', help='debug or not', action='store_true')
    parser.add_argument('--outFilterMatchNmin', help='STAR outFilterMatchNmin', default=0)
    return parser


def link_data(outdir, fq_dict):
    raw_dir = f'{outdir}/data_give/rawdata'
    os.system('mkdir -p %s' % (raw_dir))
    with open(raw_dir + '/ln.sh', 'w') as fh:
        fh.write('cd %s\n' % (raw_dir))
        for s, arr in fq_dict.items():
            fh.write('ln -sf %s %s\n' % (arr[0], s + '_1.fq.gz'))
            fh.write('ln -sf %s %s\n' % (arr[1], s + '_2.fq.gz'))


def generic_open(file_name, mode='rt'):
    if file_name.endswith('.gz'):
        file_obj = gzip.open(file_name, mode)
    else:
        file_obj = open(file_name, mode)
    return file_obj


def gene_convert(gtf_file):

    gene_id_pattern = re.compile(r'gene_id "(\S+)";')
    gene_name_pattern = re.compile(r'gene_name "(\S+)"')
    id_name = {}
    c = Counter()
    with generic_open(gtf_file) as f:
        for line in f:
            if not line.strip():
                continue
            if line.startswith('#'):
                continue
            tabs = line.split('\t')
            gtf_type, attributes = tabs[2], tabs[-1]
            if gtf_type == 'gene':
                gene_id = gene_id_pattern.findall(attributes)[-1]
                gene_names = gene_name_pattern.findall(attributes)
                if not gene_names:
                    gene_name = gene_id 
                else:
                    gene_name = gene_names[-1]
                c[gene_name] += 1
                if c[gene_name] > 1:
                    gene_name = f'{gene_name}_{c[gene_name]}'
                id_name[gene_id] = gene_name
    return id_name


@add_log
def process_read(
    read2_file, pattern_dict, barcode_dict, linker_dict,
    barcode_length, linker_length):

    # if valid, return (True)
    metrics = defaultdict(int)
    res_dict = genDict(dim=3)
    read2 = xopen.xopen(read2_file, "rt")
    while True:
        line1 = read2.readline()
        line2 = read2.readline()
        line3 = read2.readline()
        line4 = read2.readline()
        if not line4:
            break
        metrics['Total Reads'] += 1
        attr = str(line1).strip("@").split("_")
        barcode = str(attr[0])
        umi = str(attr[1])
        seq = line2.strip()
        if linker_length != 0:
            seq_linker = seq_ranges(seq, pattern_dict['L'])
            if len(seq_linker) < linker_length:
                metrics['Reads Unmapped too Short'] += 1
                continue
        if barcode_dict:
            seq_barcode = seq_ranges(seq, pattern_dict['C'])
            if barcode_length != len(seq_barcode):
                miss_length = barcode_length - len(seq_barcode)
                if miss_length > 2:
                    metrics['Reads Unmapped too Short'] += 1
                    continue
                seq_barcode = seq_barcode + "A" * miss_length                    
        
        # check linker
        if linker_length != 0:
            valid_linker = False
            for linker_name in linker_dict:
                if hamming_correct(linker_dict[linker_name], seq_linker):
                    valid_linker = True
                    break
        else:
            valid_linker = True
            
        if not valid_linker:
            metrics['Reads Unmapped Invalid Linker'] += 1
            continue

        # check barcode
        valid_barcode = False
        for barcode_name in barcode_dict:
            if hamming_correct(barcode_dict[barcode_name], seq_barcode):
                res_dict[barcode][barcode_name][umi] += 1
                valid_barcode = True
                break

        if not valid_barcode:
            metrics['Reads Unmapped Invalid Barcode'] += 1
            continue

        # mapped
        metrics['Reads Mapped'] += 1
        if metrics['Reads Mapped'] % 1000000 == 0:
            process_read.logger.info(str(metrics['Reads Mapped']) + " reads done.")

    return res_dict, metrics


def seq_ranges_exception(seq, pattern_dict):
    # get subseq with intervals in arr and concatenate
    length = len(seq)
    for x in pattern_dict:
        if length < x[1]:
            raise Exception(f'invalid seq range {x[0]}:{x[1]} in read')
    return ''.join([seq[x[0]:x[1]]for x in pattern_dict])


def seq_ranges(seq, pattern_dict):
    # get subseq with intervals in arr and concatenate
    return ''.join([seq[x[0]:x[1]]for x in pattern_dict])


def read_one_col(file):
    df = pd.read_csv(file, header=None)
    col1 = list(df.iloc[:, 0])
    num = len(col1)
    return col1, num


def read_fasta(fasta_file, equal=False):
    # seq must have equal length
    dict = {}
    LENGTH = 0
    with open(fasta_file, "rt") as f:
        while True:
            line1 = f.readline()
            line2 = f.readline()
            if not line2:
                break
            name = line1.strip(">").strip()
            seq = line2.strip()
            seq_length = len(seq)
            if equal:
                if LENGTH == 0:
                    LENGTH = seq_length
                else:
                    if LENGTH != seq_length:
                        raise Exception(f"{fasta_file} have different seq length")
            dict[name] = seq
    if equal:
        return dict, LENGTH
    return dict


def hamming_correct(string1, string2):
    threshold = len(string1) / 10 + 1
    if hamming_distance(string1, string2) < threshold:
        return True
    return False


def hamming_distance(string1, string2):
    distance = 0
    length = len(string1)
    length2 = len(string2)
    if (length != length2):
        raise Exception(f"string1({length}) and string2({length2}) do not have same length")
    for i in range(length):
        if string1[i] != string2[i]:
            distance += 1
    return distance


def gen_stat(df, stat_file):
    # 3cols: item count total_count

    def add_percent(row):
        count = row['count']
        percent = count / row['total_count']
        value = f'{format_number(count)}({round(percent * 100, 2)}%)'
        return value

    df.loc[:,'value'] = df.loc[:,'count']
    df.loc[~df['total_count'].isna(), 'value'] = df.loc[~df['total_count'].isna(), :].apply(
        lambda row: add_percent(row), axis=1
    )
    df.loc[df['total_count'].isna(), 'value'] = df.loc[df['total_count'].isna(), :].apply(
        lambda row: f'{format_number(row["count"])}', axis=1
    )
    df = df.loc[:, ["item", "value"]]
    df.to_csv(stat_file, sep=":", header=None, index=False)


def get_read(library_id, library_path, read='1'):
    read1_list = [f'_{read}', f'R{read}', f'R{read}_001']
    fq_list = ['fq', 'fastq']
    suffix_list = ["", ".gz"]
    read_pattern_list = [
        f'{library_path}/*{library_id}*{read}.{fq_str}{suffix}' 
        for read in read1_list 
        for fq_str in fq_list 
        for suffix in suffix_list
    ]
    fq_list = [glob.glob(read1_pattern) for read1_pattern in read_pattern_list]
    fq_list = sorted(non_empty for non_empty in fq_list if non_empty)
    fq_list = list(itertools.chain(*fq_list))
    if len(fq_list) == 0:
        print("Allowed R1 patterns:")
        for pattern in read_pattern_list:
            print(pattern)
        raise Exception(
            '\n'
            f'Invalid Read{read} path! \n'
            f'library_id: {library_id}\n'
            f'library_path: {library_path}\n'
        )
    return fq_list


def get_fq(library_id, library_path):
    fq1_list = get_read(library_id, library_path, read='1')
    fq2_list = get_read(library_id, library_path, read='2')
    if len(fq1_list) != len(fq2_list):
        raise Exception("Read1 and Read2 fastq number do not match!")
    fq1 = ",".join(fq1_list)
    fq2 = ",".join(fq2_list)
    return fq1, fq2


def parse_map_col4(mapfile, default_val):
    fq_dict = defaultdict(list)
    col4_dict = defaultdict(list)
    with open(mapfile) as fh:
        for line in fh:
            line = line.strip()
            if not line:
                continue
            if line.startswith('#'):
                continue
            tmp = line.split()
            library_id = tmp[0]
            library_path = tmp[1]
            sample_name = tmp[2]
            if len(tmp) == 4:
                col4 = tmp[3]
            else:
                col4 = default_val
            fq1, fq2 = get_fq(library_id, library_path)

            if sample_name in fq_dict:
                fq_dict[sample_name][0].append(fq1)
                fq_dict[sample_name][1].append(fq2)
            else:
                fq_dict[sample_name] = [[fq1], [fq2]]
            if col4 and col4 != default_val:
                col4_dict[sample_name] = col4

    for sample_name in fq_dict:
        fq_dict[sample_name][0] = ",".join(fq_dict[sample_name][0])
        fq_dict[sample_name][1] = ",".join(fq_dict[sample_name][1])

    if not fq_dict:
        raise Exception('empty mapfile!')
    return fq_dict, col4_dict


def generate_sjm(cmd, name, conda, m=1, x=1, steps_run='all'):
    res_cmd = f'''
job_begin
    name {name}
    sched_options -w n -cwd -V -l vf={m}g,p={x}
    cmd source activate {conda}; {cmd}
job_end
'''

    return res_cmd


def merge_report(
    fq_dict, steps, last_step, sjm_cmd,
    sjm_order, logdir, conda, outdir, rm_files):    
    step = "merge_report"
    steps_str = ",".join(steps)
    samples = ','.join(fq_dict.keys())
    app = tools_dir + '/merge_table.py'
    cmd = (
        f'python {app} --samples {samples} '
        f'--steps {steps_str} --outdir {outdir}'
    )
    if rm_files:
        cmd += ' --rm_files'
    sjm_cmd += generate_sjm(cmd, 'merge_report', conda)
    for sample in fq_dict:
        sjm_order += f'order {step} after {last_step}_{sample}\n'
    with open(logdir + '/sjm.job', 'w') as fh:
        fh.write(sjm_cmd + '\n')
        fh.write(sjm_order)


def format_number(number: int) -> str:
    return format(number, ",")


def format_metrics(metrics: dict):
    for key in metrics:
        value = metrics[key]
        metrics[key] = format_number(value)
        if isinstance(value, float):
            metrics[key] = round(value, 2)


def format_ratios(ratios: dict):
    for key in ratios:
        ratios[key] = round(ratios[key] * 100, 2)


@add_log
def glob_genomeDir(genomeDir, fa=False):
    refFlat = glob.glob(genomeDir + "/*.refFlat")
    if (len(refFlat) > 1):
        sys.exit("ERROR: Multiple refFlat file in " + genomeDir)
    elif (len(refFlat) == 0):
        sys.exit("ERROR: refFlat file not found in " + genomeDir)
    else:
        refFlat = refFlat[0]
        glob_genomeDir.logger.info("refFlat file found: " + refFlat)

    gtf = glob.glob(genomeDir + "/*.gtf")
    if (len(gtf) == 0):
        sys.exit("ERROR: gtf file not found in " + genomeDir)
    elif (len(gtf) > 1):
        gtf = glob.glob(genomeDir + "/*.chr.gtf")
        if (len(gtf) == 0):
            sys.exit("ERROR: No chr gtf file in "+ genomeDir)
        if (len(gtf) > 1):
            sys.exit("ERROR: Multiple gtf file in " + genomeDir)
        else:
            gtf = gtf[0]
            glob_genomeDir.logger.info("chr gtf file found: " + gtf)
    else:
        gtf = gtf[0]
        glob_genomeDir.logger.info("gtf file found: " + gtf)
    
    if fa:
        fasta = glob.glob(genomeDir + "/*.fa") + glob.glob(genomeDir + "/*.fasta")
        if len(fasta) > 1:
            sys.exit("ERROR: Multiple fasta file in " + genomeDir)
        fasta = fasta[0]
        return refFlat, gtf, fasta
    return refFlat, gtf


def barcode_filter_with_magnitude(
        df, plot='magnitude.pdf', col='UMI', percent=0.1, expected_cell_num=3000):
    # col can be readcount or UMI
    # determine validated barcodes
    df = df.sort_values(col, ascending=False)
    idx = int(expected_cell_num * 0.01) - 1
    idx = max(0, idx)

    # calculate read counts threshold
    threshold = int(df.iloc[idx, df.columns == col] * 0.1)
    threshold = max(1, threshold)
    validated_barcodes = df[df[col] > threshold].index

    fig = plt.figure()
    plt.plot(df[col])
    plt.hlines(threshold, 0, len(validated_barcodes), linestyle='dashed')
    plt.vlines(len(validated_barcodes), 0, threshold, linestyle='dashed')
    plt.title('expected cell num: %s\n%s threshold: %s\ncell num: %s' %
              (expected_cell_num, col, threshold, len(validated_barcodes)))
    plt.loglog()
    plt.savefig(plot)

    return (validated_barcodes, threshold, len(validated_barcodes))


def barcode_filter_with_kde(df, plot='kde.pdf', col='UMI'):
    # col can be readcount or UMI
    # filter low values
    df = df.sort_values(col, ascending=False)
    arr = np.log10([i for i in df[col] if i / float(df[col][0]) > 0.001])

    # kde
    x_grid = np.linspace(min(arr), max(arr), 10000)
    density = gaussian_kde(arr, bw_method=0.1)
    y = density(x_grid)

    local_mins = argrelextrema(y, np.less)
    log_threshold = x_grid[local_mins[-1][0]]
    threshold = np.power(10, log_threshold)
    validated_barcodes = df[df[col] > threshold].index

    # plot
    fig, (ax1, ax2) = plt.subplots(2, figsize=(6.4, 10))
    ax1.plot(x_grid, y)
    #ax1.axhline(y[local_mins[-1][0]], -0.5, log_threshold, linestyle='dashed')
    ax1.vlines(log_threshold, 0, y[local_mins[-1][0]], linestyle='dashed')
    ax1.set_ylim(0, 0.3)

    ax2.plot(df[col])
    ax2.hlines(threshold, 0, len(validated_barcodes), linestyle='dashed')
    ax2.vlines(len(validated_barcodes), 0, threshold, linestyle='dashed')
    ax2.set_title('%s threshold: %s\ncell num: %s' %
                  (col, int(threshold), len(validated_barcodes)))
    ax2.loglog()
    plt.savefig(plot)

    return (validated_barcodes, threshold, len(validated_barcodes))


def get_slope(x, y, window=200, step=10):
    assert len(x) == len(y)
    start = 0
    last = len(x)
    res = [[], []]
    while True:
        end = start + window
        if end > last:
            break
        z = np.polyfit(x[start:end], y[start:end], 1)
        res[0].append(x[start])
        res[1].append(z[0])
        start += step
    return res


def barcode_filter_with_derivative(
        df, plot='derivative.pdf', col='UMI', window=500, step=5):
    # col can be readcount or UMI
    # filter low values
    df = df.sort_values(col, ascending=False)
    y = np.log10([i for i in df[col] if i / float(df[col][0]) > 0.001])
    x = np.log10(np.arange(len(y)) + 1)

    # derivative
    res = get_slope(x, y, window=window, step=step)
    res2 = get_slope(res[0], res[1], window=window, step=step)
    g0 = [res2[0][i] for i, j in enumerate(res2[1]) if j > 0]
    cell_num = int(np.power(10, g0[0]))
    threshold = df[col][cell_num]
    validated_barcodes = df.index[0:cell_num]

    # plot
    fig, (ax1, ax2, ax3) = plt.subplots(3, figsize=(6.4, 15))
    ax1.plot(res[0], res[1])

    ax2.plot(res2[0], res2[1])
    ax2.set_ylim(-1, 1)

    ax3.plot(df[col])
    ax3.hlines(threshold, 0, len(validated_barcodes), linestyle='dashed')
    ax3.vlines(len(validated_barcodes), 0, threshold, linestyle='dashed')
    ax3.set_title('%s threshold: %s\ncell num: %s' %
                  (col, int(threshold), len(validated_barcodes)))
    ax3.loglog()
    plt.savefig(plot)

    return (validated_barcodes, threshold, len(validated_barcodes))


def genDict(dim=3, valType=int):
    if dim == 1:
        return defaultdict(valType)
    else:
        return defaultdict(lambda: genDict(dim - 1, valType=valType))

'''
@add_log
def downsample(bam, barcodes, percent):
    """
    calculate median_geneNum and saturation based on a given percent of reads

    Args:
        bam - bam file
        barcodes(set) - validated barcodes
        percent(float) - percent of reads in bam file.

    Returns:
        percent(float) - input percent
        median_geneNum(int) - median gene number
        saturation(float) - sequencing saturation

    Description:
        Sequencing Saturation = 1 - n_deduped_reads / n_reads.
        n_deduped_reads = Number of unique (valid cell-barcode, valid UMI, gene) combinations among confidently mapped reads.
        n_reads = Total number of confidently mapped, valid cell-barcode, valid UMI reads.
    """
    logging.info('working' + str(percent))
    cmd = ['samtools', 'view', '-s', str(percent), bam]
    p1 = subprocess.Popen(cmd, stdout=subprocess.PIPE)
    # nesting defaultdicts in an arbitrary depth
    readDict = genDict()

    n_reads = 0
    while True:
        line = p1.stdout.readline()
        if not line.strip():
            break
        tmp = line.strip().split()
        if tmp[-1].startswith('XT:Z:'):
            geneID = tmp[-1].replace('XT:Z:', '')
            cell_barcode, umi = tmp[0].split('_')[0:2]
            # filter invalid barcode
            if cell_barcode in barcodes:
                n_reads += 1
                readDict[cell_barcode][umi][geneID] += 1
    p1.stdout.close()

    geneNum_list = []
    n_deduped_reads = 0
    for cell_barcode in readDict:
        genes = set()
        for umi in readDict[cell_barcode]:
            for geneID in readDict[cell_barcode][umi]:
                genes.add(geneID)
                if readDict[cell_barcode][umi][geneID] == 1:
                    n_deduped_reads += 1
        geneNum_list.append(len(genes))

    median_geneNum = np.median(geneNum_list) if geneNum_list else 0
    saturation = (1 - float(n_deduped_reads) / n_reads) * 100

    return "%.2f\t%.2f\t%.2f\n" % (
        percent, median_geneNum, saturation), saturation
'''


def cluster_tsne_list(tsne_df):
    """
    tSNE_1	tSNE_2	cluster Gene_Counts
    return data list
    """
    sum_df = tsne_df.groupby(["cluster"]).agg("count").iloc[:, 0]
    percent_df = sum_df.transform(lambda x: round(x / sum(x) * 100, 2))
    res = []
    for cluster in sorted(tsne_df.cluster.unique()):
        sub_df = tsne_df[tsne_df.cluster == cluster]
        name = "cluster {cluster}({percent}%)".format(
            cluster=cluster, percent=percent_df[cluster])
        tSNE_1 = list(sub_df.tSNE_1)
        tSNE_2 = list(sub_df.tSNE_2)
        res.append({"name": name, "tSNE_1": tSNE_1, "tSNE_2": tSNE_2})
    return res


def marker_table(marker_df):
    """
    return html code
    """
    marker_df = marker_df.loc[:, ["cluster", "gene",
                                  "avg_logFC", "pct.1", "pct.2", "p_val_adj"]]
    marker_gene_table = marker_df.to_html(
        escape=False,
        index=False,
        table_id="marker_gene_table",
        justify="center")
    return marker_gene_table


def report_prepare(outdir, **kwargs):
    json_file = outdir + '/../.data.json'
    if not os.path.exists(json_file):
        data = {}
    else:
        fh = open(json_file)
        data = json.load(fh)
        fh.close()

    for key in kwargs:
        data[key] = kwargs[key]

    with open(json_file, 'w') as fh:
        json.dump(data, fh)


def parse_vcf(vcf_file, cols=['chrom', 'pos', 'alleles'], infos=['VID','CID']):
    vcf = pysam.VariantFile(vcf_file)
    df = pd.DataFrame(columns=[col.capitalize() for col in cols] + infos)
    rec_dict = {}
    for rec in vcf.fetch():

        for col in cols:
            rec_dict[col.capitalize()] = getattr(rec, col)
            if col == 'alleles':
                rec_dict['Alleles'] = '-'.join(rec_dict['Alleles'])
                
        for info in infos:
            rec_dict[info] = rec.info[info]

        '''
        rec_dict['GT'] = [s['GT'] for s in rec.samples.values()][0]
        rec_dict['GT'] = [str(item) for item in rec_dict['GT']]
        rec_dict['GT'] = '/'.join(rec_dict['GT'])
        '''

        df = df.append(pd.Series(rec_dict),ignore_index=True)
    return df


def parse_annovar(annovar_file):
    df = pd.DataFrame(columns=['Gene','mRNA', 'Protein', 'COSMIC'])
    with open(annovar_file, 'rt') as f:
        index = 0
        for line in f:
            index += 1
            if index == 1:
                continue
            attrs = line.split('\t')
            gene = attrs[6]
            func = attrs[5]
            if func == 'exonic':
                changes = attrs[9]
                cosmic = attrs[10]
            else:
                changes = attrs[7]
                cosmic = attrs[8]
            change_list = list()
            for change in changes.split(','):
                change_attrs = change.split(':')
                mRNA = ''
                protein = ''
                for change_index in range(len(change_attrs)):
                    change_attr = change_attrs[change_index]
                    if change_attr.startswith('c.'):
                        base = change_attr.strip('c.')
                        exon = change_attrs[change_index - 1]
                        mRNA = f'{exon}:{base}'
                    if change_attr.startswith('p.'):
                        protein = change_attr.strip('p.')
                if not (mRNA, protein) in change_list:
                    change_list.append((mRNA, protein)) 
            combine = [','.join(item) for item in list(zip(*change_list))]
            mRNA = combine[0]
            protein = combine[1]
            df = df.append({
                'Gene': gene,
                'mRNA': mRNA,
                'Protein': protein,
                'COSMIC': cosmic,
            }, ignore_index=True)
    return df


def parse_match_dir(match_dir):
    match_dict = {}
    match_barcode, cell_total = read_barcode_file(match_dir)
    match_dict['match_barcode'] = match_barcode
    match_dict['cell_total'] = cell_total
    match_dict['tsne_coord'] = glob.glob(f'{match_dir}/*analysis*/*tsne_coord.tsv')[0]
    match_dict['markers'] = glob.glob(f'{match_dir}/*analysis*/*markers.tsv')[0]
    try:
        match_dict['rds'] = glob.glob(f'{match_dir}/*analysis/*.rds')[0]
    except Exception:
        match_dict['rds'] = None
    return match_dict


@add_log
def STAR_util(
    sample,
    outdir,
    input_read,
    genomeDir,
    runThreadN,
    outFilterMatchNmin=35,
    out_unmapped=False,
    outFilterMultimapNmax=1,
    outBAMsortingBinsN=50,
):

    # check dir
    if not os.path.exists(outdir):
        os.system('mkdir -p %s' % (outdir))

    out_prefix = f'{outdir}/{sample}_'
    out_BAM = out_prefix + "Aligned.sortedByCoord.out.bam"

    cmd = f"STAR \
 --genomeDir {genomeDir} \
 --readFilesIn {input_read}\
 --readFilesCommand zcat\
 --outSAMtype BAM SortedByCoordinate\
 --runThreadN {runThreadN}\
 --outFilterMatchNmin {outFilterMatchNmin}\
 --outFileNamePrefix {out_prefix}\
 --outFilterMultimapNmax {outFilterMultimapNmax}\
 --outBAMsortingBinsN {outBAMsortingBinsN}\
 --limitBAMsortRAM 100000000000 "
    
    if out_unmapped:
        cmd += ' --outReadsUnmapped Fastx '

    STAR_util.logger.info(cmd)
    os.system(cmd)

    cmd = "samtools index {out_BAM}".format(out_BAM=out_BAM)
    STAR_util.logger.info(cmd)
    os.system(cmd)


def get_scope_bc(bctype):
    import celescope
    root_path = os.path.dirname(celescope.__file__)
    linker_f = glob.glob(f'{root_path}/data/chemistry/{bctype}/linker*')[0]
    whitelist_f = f'{root_path}/data/chemistry/{bctype}/bclist'
    return linker_f, whitelist_f

@add_log
def parse_pattern(pattern):
    # 解析接头结构，返回接头结构字典
    # key: 字母表示的接头, value: 碱基区间列表
    # eg.: C8L10C8L10C8U8T30
    # defaultdict(<type 'list'>:
    # {'C': [[0, 8], [18, 26], [36, 44]], 'U': [[44, 52]], 'L': [[8, 18], [26, 36]], 'T': [[52, 82]]})
    pattern_dict = defaultdict(list)
    p = re.compile(r'([CLUNT])(\d+)')
    tmp = p.findall(pattern)
    if not tmp:
        parse_pattern.logger.error(f'Invalid pattern: {pattern}')
        sys.exit()
    start = 0
    for item in tmp:
        end = start + int(item[1])
        pattern_dict[item[0]].append([start, end])
        start = end
    return pattern_dict
    

def fastq_line(name, seq, qual):
    return f'@{name}\n{seq}\n+\n{qual}\n'


def s_common(parser):
    """subparser common arguments
    """
    parser.add_argument('--outdir', help='output dir', required=True)
    parser.add_argument('--assay', help='assay', required=True)
    parser.add_argument('--sample', help='sample name', required=True)
    parser.add_argument('--thread', default=4)
    parser.add_argument('--debug', help='debug', action='store_true')
    return parser


def find_assay_init(assay):
    init_module = importlib.import_module(f"celescope.{assay}.__init__")
    return init_module


def find_step_module(assay, step):
    init_module = find_assay_init(assay)
    try:
        step_module = importlib.import_module(f"celescope.{assay}.{step}")
    except ModuleNotFoundError as error:
        try:
            step_module = importlib.import_module(f"celescope.tools.{step}")
        except ModuleNotFoundError as error:
            module_path = init_module.IMPORT_DICT[step]
            step_module = importlib.import_module(f"{module_path}.{step}")

    return step_module


import os
import glob
import sys
import argparse
import re
from collections import defaultdict
from celescope.hla.__init__ import __STEPS__, __ASSAY__
from celescope.tools.utils import *


def main():

    # init
    assay = __ASSAY__
    steps = __STEPS__
    conda = os.environ['CONDA_DEFAULT_ENV']
    app = 'celescope'

    # parser
    parser = multi_opts(assay)
    parser.add_argument('--thread', help='thread', default=6)
    args = parser.parse_args()

    # read args
    outdir = args.outdir
    chemistry = args.chemistry
    pattern = args.pattern
    whitelist = args.whitelist
    linker = args.linker
    lowQual = args.lowQual
    lowNum = args.lowNum
    mod = args.mod
    rm_files = args.rm_files

    # parse mapfile
    fq_dict, match_dict = parse_map_col4(args.mapfile, None)

    # link
    link_data(outdir, fq_dict)

    # custom args
    thread = args.thread

    # mk log dir
    logdir = outdir + '/log'
    os.system('mkdir -p %s' % (logdir))

    # script init
    sjm_cmd = 'log_dir %s\n' % (logdir)
    sjm_order = ''
    shell = ''

    # outdir dict
    for sample in fq_dict:
        outdir_dic = {}
        index = 0
        for step in steps:
            step_outdir = f"{outdir}/{sample}/{index:02d}.{step}"
            outdir_dic.update({step: step_outdir})
            index += 1

        # sample
        step = "sample"
        cmd = (
            f'{app} {assay} {step} '
            f'--chemistry {chemistry} '
            f'--sample {sample} --outdir {outdir_dic[step]} --assay {assay} '
        )
        sjm_cmd += generate_sjm(cmd, f'{step}_{sample}', conda)
        shell += cmd + '\n'
        last_step = step

        # barcode
        arr = fq_dict[sample]
        step = "barcode"
        cmd = (
            f'{app} {assay} {step} '
            f'--fq1 {arr[0]} --fq2 {arr[1]} --chemistry {chemistry} '
            f'--pattern {pattern} --whitelist {whitelist} --linker {linker} '
            f'--sample {sample} --lowQual {lowQual} --thread {thread} '
            f'--lowNum {lowNum} --outdir {outdir_dic[step]} --assay {assay} '

        )
        sjm_cmd += generate_sjm(cmd, f'{step}_{sample}', conda, m=5, x=thread)
        sjm_order += f'order {step}_{sample} after {last_step}_{sample}\n'
        shell += cmd + '\n'
        last_step = step

        # adapt
        step = "cutadapt"
        fq = f'{outdir_dic["barcode"]}/{sample}_2.fq'
        cmd = (
            f'{app} {assay} {step} '
            f'--fq {fq} --sample {sample} --outdir '
            f'{outdir_dic[step]} --assay {assay} '
        )
        sjm_cmd += generate_sjm(cmd, f'{step}_{sample}', conda, m=5, x=1)
        sjm_order += f'order {step}_{sample} after {last_step}_{sample}\n'
        shell += cmd + '\n'
        last_step = step

        # mapping_hla
        step = 'mapping_hla'
        fq = f'{outdir_dic["cutadapt"]}/{sample}_clean_2.fq'
        cmd = (
            f'{app} {assay} {step} '
            f'--fq {fq} --sample {sample} '
            f'--thread {thread} '
            f'--match_dir {match_dict[sample]} '
            f'--outdir {outdir_dic[step]} --assay {assay} '
        )
        sjm_cmd += generate_sjm(cmd, f'{step}_{sample}', conda, m=30, x=thread)
        sjm_order += f'order {step}_{sample} after {last_step}_{sample}\n'
        shell += cmd + '\n'
        last_step = step

     # merged report
    if mod == 'sjm':
        step = 'merge_report'
        merge_report(
            fq_dict,
            steps,
            last_step,
            sjm_cmd,
            sjm_order,
            logdir,
            conda,
            outdir,
            rm_files,
        )
    if mod == 'shell':
        os.system('mkdir -p ./shell/')
        with open(f'./shell/{sample}.sh', 'w') as f:
            f.write(shell)


if __name__ == '__main__':
    main()



import unittest
import os
import pandas as pd
import glob
from celescope.snp.snpCalling import *
from celescope.tools.utils import read_barcode_file, glob_genomeDir, parse_vcf
from .analysis_snp import analysis_variant


class test_snp(unittest.TestCase):
    '''
    def setUp(self):
        os.chdir('/SGRNJ01/RD_dir/pipeline_test/zhouyiqi/unittest/snp')
        self.genomeDir = '/SGRNJ/Public/Database/genome/homo_sapiens/ensembl_92'
        self.gene_list_file = './gene_list.tsv'
        self.index_file = './S20070818_TS/05.snpCalling/S20070818_TS_cell_index.tsv'
        self.thread = 20
        self.outdir = './S20070818_TS/05.snpCalling/'
        self.analysis_outdir = './S20070818_TS/06.analysis_snp'
        _refFlat, self.gtf, self.fasta = glob_genomeDir(self.genomeDir, fa=True)
        self.match_dir = '/SGRNJ02/RandD4/RD20051303_Panel/20200717/S20070818_ZL/'
        self.sample = 'S20070818_TS'
        self.count_file = './S20070818_TS/05.snpCalling/S20070818_TS_count.tsv'
        self.vcf_file = './S20070818_TS/05.snpCalling/S20070818_TS_anno.vcf'
        self.assay = 'snp'
        self.annovar_config = '/SGRNJ01/RD_dir/pipeline_test/zhouyiqi/soft/annovar/annovar.config'
        self.step_analysis_variant = analysis_variant(
            self.analysis_outdir,
            self.sample,
            self.match_dir,
            self.vcf_file,
            self.index_file,
            self.assay,
            self.annovar_config,
        )
    '''

    def setUp(self):
        os.chdir('/SGRNJ01/RD_dir/pipeline_test/zhouyiqi/unittest/snp')
        sample = '20201224_consplit'
        self.sample = sample
        self.genomeDir = '/SGRNJ/Public/Database/genome/homo_sapiens/ensembl_92'
        self.gene_list_file = './gene_list.tsv'
        self.CID_file = f'./{sample}/05.snpCalling/{sample}_CID.tsv'
        self.thread = 20
        self.analysis_outdir = f'./{sample}/06.analysis_snp'
        _refFlat, self.gtf, self.fasta = glob_genomeDir(self.genomeDir, fa=True)
        self.match_dir = '/SGRNJ02/RandD4/RD20051303_Panel/20200717/S20070818_ZL/'
        self.vcf_file = f'./{sample}/05.snpCalling/{sample}_merged.vcf'
        self.assay = 'snp'
        self.annovar_config = '/SGRNJ01/RD_dir/pipeline_test/zhouyiqi/soft/annovar/annovar.config'
        self.step_analysis_variant = analysis_variant(
            self.analysis_outdir,
            self.sample,
            self.match_dir,
            self.vcf_file,
            self.CID_file,
            self.assay,
            self.annovar_config,
        )


    @unittest.skip('pass')
    def test_annovar(self):
        self.step_analysis_variant.annovar()
    
    @unittest.skip('pass')
    def test_parse_annovar(self):
        from celescope.tools.utils import parse_annovar
        self.annovar_file = glob.glob(f'{self.analysis_outdir}/{self.sample}*_multianno.txt')[0]
        parse_annovar(self.annovar_file)

    def test_parse_vcf(self):
        self.step_analysis_variant.get_df_vcf()
        print(self.step_analysis_variant.df_vcf)
        

if __name__ == '__main__':
    unittest.main()



import os
from collections import defaultdict
from concurrent.futures import ProcessPoolExecutor
import celescope
import pysam
import numpy as np
import pandas as pd
import logging
from celescope.tools.utils import *
from celescope.tools.report import reporter
from mutract.utils import Mutract


@add_log
def snpCalling(args):

    sample = args.sample
    outdir = args.outdir
    thread = int(args.thread)
    match_dir = args.match_dir
    bam = args.bam
    genomeDir = args.genomeDir
    gene_list_file = args.gene_list

    # check dir
    if not os.path.exists(outdir):
        os.system('mkdir -p %s' % (outdir))

    # get args
    _refFlat, _gtf, fasta = glob_genomeDir(genomeDir, fa=True)
    _match_barcode, _cell_total, match_barcode_file = read_barcode_file(match_dir, return_file=True)

    # mutract
    obj = Mutract(
        outdir, sample, bam, fasta, 
        match_barcode_file, thread=thread, gene_file=gene_list_file
    )
    obj.run()


def get_opts_snpCalling(parser, sub_program):
    if sub_program:
        s_common(parser)
        parser.add_argument("--bam", help='featureCounts bam', required=True)
        parser.add_argument(
        "--match_dir", help="match scRNA-Seq dir", required=True)
    parser.add_argument("--genomeDir", help='genomeDir', required=True)
    parser.add_argument("--gene_list", help='gene_list', required=True)



import itertools
import pysam
import collections
import numpy as np

import celescope.tools.utils as utils
from celescope.tools.Step import Step, s_common


@utils.add_log
def target_metrics(args):
    step_name = "target_metrics"
    step = Step(args, step_name)

    gene_list, n_gene = utils.read_one_col(args.gene_list)
    step.add_metric(
        name="Number of Target Genes",
        value=n_gene,
    )

    match_barcode = set(utils.parse_match_dir(args.match_dir)["match_barcode"])

    count_dict = utils.genDict(dim=3, valType=int)
    with pysam.AlignmentFile(args.bam, "rb") as reader:
        for record in reader:
            try:
                gene_name = record.get_tag('GN')
            except KeyError:
                continue
            barcode = record.get_tag('CB')
            UMI = record.get_tag('UB')
            count_dict[barcode][gene_name][UMI] += 1

    UMIs = 0
    enriched_UMIs = 0
    enriched_UMIs_in_cells = 0
    enriched_UMIs_per_cell = []


    for barcode in count_dict:
        barcode_enriched_UMI = 0
        for gene_name in count_dict[barcode]:
            gene_UMI = len(count_dict[barcode][gene_name])
            UMIs += gene_UMI
            if gene_name in gene_list:
                enriched_UMIs += gene_UMI
                if barcode in match_barcode:
                    enriched_UMIs_in_cells += gene_UMI
                    barcode_enriched_UMI += gene_UMI
        if barcode in match_barcode:
            enriched_UMIs_per_cell.append(barcode_enriched_UMI)
    target_metrics.logger.debug(enriched_UMIs_per_cell)

    step.add_metric(
        name="Enriched UMIs",
        value=enriched_UMIs,
        total=UMIs,
    )
    step.add_metric(
        name="Enriched UMIs in Cells",
        value=enriched_UMIs_in_cells,
        total=UMIs,
    )
    step.add_metric(
        name="Median Enriched UMIs per Cell",
        value=np.median(enriched_UMIs_per_cell),
    )

    step.clean_up()


def get_opts_target_metrics(parser, sub_program):
    if sub_program:
        parser = s_common(parser)
        parser.add_argument("--bam", help='featureCounts bam', required=True)
        parser.add_argument('--match_dir', help='match_dir', required=True)
    parser.add_argument("--gene_list", help='gene_list', required=True)





from collections import defaultdict
import pysam
import gzip
import os
import pandas as pd
import logging
import numpy as np
import sys
import argparse
import matplotlib as mpl
import re
import json
import glob
mpl.use('Agg')
from matplotlib import pyplot as plt
from celescope.tools.utils import *
from celescope.tools.report import reporter
import celescope.fusion

fusionDir = os.path.dirname(celescope.fusion.__file__)


def read_pos(fusion_pos_file):
    dic = {}
    df = pd.read_csv(fusion_pos_file, sep="\t")
    for tag, pos in zip(df.iloc[:, 0], df.iloc[:, 1]):
        dic[tag] = pos
    return dic


def is_fusion(pos, read_start, read_length, flanking_base):
    test_start = (pos - flanking_base) >= read_start
    test_end = (pos + flanking_base) <= (read_start + read_length)
    return (test_start and test_end)


@add_log
def count_fusion(args):

    outdir = args.outdir
    sample = args.sample
    bam = args.bam
    flanking_base = int(args.flanking_base)
    fusion_pos_file = args.fusion_pos
    match_dir = args.match_dir
    UMI_min = int(args.UMI_min)

    # check dir
    if not os.path.exists(outdir):
        os.system('mkdir -p %s' % (outdir))

    fusion_pos = read_pos(fusion_pos_file)
    out_prefix = outdir + "/" + sample
    # barcode
    match_barcode, _n_barcode = read_barcode_file(match_dir)
    # tsne
    match_tsne_file = parse_match_dir(match_dir)['tsne_coord']
    df_tsne = pd.read_csv(match_tsne_file, sep="\t", index_col=0)
    # out
    out_read_count_file = out_prefix + "_fusion_read_count.tsv"
    out_umi_count_file = out_prefix + "_fusion_UMI_count.tsv"
    out_barcode_count_file = out_prefix + "_fusion_barcode_count.tsv"
    out_tsne_file = out_prefix + "_fusion_tsne.tsv"

    # process bam
    samfile = pysam.AlignmentFile(bam, "rb")
    header = samfile.header
    new_bam = pysam.AlignmentFile(
        out_prefix + "_fusion.bam", "wb", header=header)
    count_dic = genDict(dim=3)
    for read in samfile:
        tag = read.reference_name
        read_start = int(read.reference_start)
        read_length = len(read.query_sequence)
        attr = read.query_name.split('_')
        barcode = attr[0]
        umi = attr[1]
        if tag in fusion_pos.keys():
            if barcode in match_barcode:
                if is_fusion(pos=fusion_pos[tag], read_start=read_start,
                             read_length=read_length, flanking_base=flanking_base):
                    new_bam.write(read)
                    count_dic[barcode][tag][umi] += 1
    new_bam.close()

    # write dic to pandas df
    rows = []
    for barcode in count_dic:
        for tag in count_dic[barcode]:
            for umi in count_dic[barcode][tag]:
                rows.append([barcode, tag, umi, count_dic[barcode][tag][umi]])
    df_read = pd.DataFrame(rows)
    df_read.rename(
        columns={
            0: "barcode",
            1: "tag",
            2: "UMI",
            3: "read_count"},
        inplace=True)
    df_read.to_csv(out_read_count_file, sep="\t", index=False)

    if not rows:
        count_fusion.logger.error('***** NO FUSION FOUND! *****')
    else:
        df_umi = df_read.groupby(["barcode", "tag"]).agg({"UMI": "count"})
        df_umi = df_umi[df_umi["UMI"] >= UMI_min]
        df_umi.to_csv(out_umi_count_file, sep="\t")

        df_umi.reset_index(inplace=True)
        df_barcode = df_umi.groupby(["tag"]).agg({"barcode": "count"})
        n_match_barcode = len(match_barcode)
        # add zero count tag
        for tag in fusion_pos.keys():
            if not tag in df_barcode.barcode:
                new_row = pd.Series(data={'barcode': 0}, name=tag)
                df_barcode = df_barcode.append(new_row, ignore_index=False)
        df_barcode["percent"] = df_barcode["barcode"] / n_match_barcode
        df_barcode.to_csv(out_barcode_count_file, sep="\t")

        df_pivot = df_umi.pivot(index="barcode", columns="tag", values="UMI")
        df_pivot.fillna(0, inplace=True)
        df_tsne_fusion = pd.merge(
            df_tsne,
            df_pivot,
            right_index=True,
            left_index=True,
            how="left")
        df_tsne_fusion.fillna(0, inplace=True)
        df_tsne_fusion.to_csv(out_tsne_file, sep="\t")

        # plot
        count_fusion.logger.info("plot fusion...!")
        app = fusionDir + "/plot_fusion.R"
        cmd = f"Rscript {app} --tsne_fusion {out_tsne_file} --outdir {outdir}"
        os.system(cmd)
        count_fusion.logger.info("plot done.")


def get_opts_count_fusion(parser, sub_program):
    if sub_program:
        s_common(parser)
        parser.add_argument("--bam", required=True)
        parser.add_argument("--match_dir", help="match scRNA-Seq dir", required=True)

    #parser.add_argument("--fusion_fasta",help="fusion fasta",required=True)
    parser.add_argument(
        "--fusion_pos",
        help="first base position of the second gene(0-start),tsv file",
        required=True)
    parser.add_argument("--flanking_base", default=5)
    parser.add_argument("--UMI_min", default=1)



import os
import subprocess
from celescope.tools.utils import *
from celescope.tools.STAR import Step_mapping

@add_log
def STAR_virus(args):

    # check dir
    if not os.path.exists(args.outdir):
        os.system('mkdir -p %s' % (args.outdir))

    sample = f'{args.sample}_virus'
    mapping = Step_mapping(
        sample, 
        args.outdir, 
        args.assay, 
        args.thread,
        args.input_read, 
        args.virus_genomeDir,
        outFilterMultimapNmax=10,
        outFilterMatchNmin=args.outFilterMatchNmin,
        sort_BAM=False
    )
    mapping.STAR()
    mapping.sort_bam()
    mapping.index_bam()


def get_opts_STAR_virus(parser, sub_program):
    if sub_program:
        s_common(parser)
        parser.add_argument("--input_read", required=True)

    parser.add_argument(
        '--virus_genomeDir',
        help='virus genome dir',
        required=True)
    parser.add_argument("--outFilterMatchNmin", help='STAR outFilterMatchNmin', default=35)
    parser.add_argument('--starMem', help='starMem', default=30)



#!/bin/env python
# coding=utf8

import os
import re
import logging
import subprocess
import glob
import sys
import pysam
from celescope.tools.utils import *
from celescope.tools.report import reporter
from celescope.tools.Step import Step


def format_stat(log, samplename):
    #Assigned, Unassigned_NoFeatures, Unassigned_Ambiguity=(0, 0, 0)
    tmp_arr = []
    fh = open(log, 'r')
    with open(os.path.dirname(log) + '/stat.txt', 'w') as stat_fh:
        p1 = re.compile(r'Assigned.*?(\d+)', flags=re.S)
        p2 = re.compile(r'Unassigned_NoFeatures.*?(\d+)', flags=re.S)
        p3 = re.compile(r'Unassigned_Ambiguity.*?(\d+)', flags=re.S)
        for line in fh:
            if line.strip() == '':
                continue

            m1 = p1.search(line.strip())
            if m1:
                tmp_arr.append(int(m1.group(1)))

            m2 = p2.search(line)
            if m2:
                tmp_arr.append(int(m2.group(1)))

            m3 = p3.search(line)
            if m3:
                tmp_arr.append(int(m3.group(1)))

        total = sum(tmp_arr)
        tmp_arr = [
            '%s(%.2f%%)' %
            (format_number(n), (n + 0.0) / total * 100) for n in tmp_arr]
        for t, s in zip(['Assigned', 'Unassigned_NoFeatures',
                         'Unassigned_Ambiguity'], tmp_arr):
            stat_fh.write('%s: %s\n' % (t, s))
    fh.close()

@add_log
def add_tag(bam, gtf):
    id_name = gene_convert(gtf)
    samfile = pysam.AlignmentFile(bam, "rb")
    header = samfile.header
    new_bam = pysam.AlignmentFile(
        bam + ".temp", "wb", header=header)
    for read in samfile:
        attr = read.query_name.split('_')
        barcode = attr[0]
        umi = attr[1]
        read.set_tag(tag='CB', value=barcode, value_type='Z')
        read.set_tag(tag='UB', value=umi, value_type='Z')
        if read.has_tag('XT'):
            gene_id = read.get_tag('XT')
            gene_name = id_name[gene_id]
            read.set_tag(tag='GN', value=gene_name, value_type='Z')
            read.set_tag(tag='GX', value=gene_id, value_type='Z')
        new_bam.write(read)
    new_bam.close()
    cmd = f'mv {bam}.temp {bam}'
    subprocess.check_call(cmd, shell=True)

@add_log
def featureCounts(args):

    step_name = "featureCounts"
    step = Step(args, step_name)

    # check
    if args.genomeDir and args.genomeDir != "None":
        _refFlat, gtf = glob_genomeDir(args.genomeDir)
    else:
        gtf = args.gtf

    # run featureCounts
    outPrefix = args.outdir + '/' + args.sample
    cmd = [
        'featureCounts', 
        '-s', '1',
        '-a', gtf, 
        '-o', outPrefix, 
        '-R', 'BAM',
        '-T', str(args.thread), 
        '-t', args.gtf_type, 
        args.input,
    ]
    featureCounts.logger.info('%s' % (' '.join(cmd)))
    subprocess.check_call(cmd)

    subprocess.check_call(['which', 'samtools'])

    bam_basename = os.path.basename(args.input)
    bam = f'{args.outdir}/{bam_basename}.featureCounts.bam'

    # add tag
    add_tag(bam, gtf)

    # sort by name:BC and umi
    featureCounts.logger.info('samtools sort ...!')
    cmd = [
        'samtools',
        'sort',
        '-n',
        '-@',
        str(args.thread),
        '-o',
        outPrefix +
        '_name_sorted.bam',
        bam]
    featureCounts.logger.info('%s' % (' '.join(cmd)))
    subprocess.check_call(cmd)
    featureCounts.logger.info('samtools sort done.')

    format_stat(args.outdir + '/' + args.sample + '.summary', args.sample)

    step.clean_up()


def get_opts_featureCounts(parser, sub_program):
    parser.add_argument('--gtf', help='gtf file path')
    parser.add_argument(
        '--gtf_type',
        help='Specify feature type in GTF annotation',
        default='exon')
    parser.add_argument('--genomeDir')
    if sub_program:
        parser = s_common(parser)
        parser.add_argument('--input', required=True)
    return parser




#!/bin/env python
# coding=utf8

import os
import sys
import json
import logging
import re
import numpy as np
import pandas as pd
import glob
from scipy.io import mmwrite
from scipy.sparse import csr_matrix
from celescope.tools.report import reporter
from celescope.tools.utils import *
import celescope.tools

toolsdir = os.path.dirname(celescope.tools.__file__)


def report_prepare(outdir, tsne_df, marker_df, virus_df):
    json_file = outdir + '/../.data.json'
    if not os.path.exists(json_file):
        data = {}
    else:
        fh = open(json_file)
        data = json.load(fh)
        fh.close()

    data["cluster_tsne"] = cluster_tsne_list(tsne_df)
    data["virus_tsne"] = virus_tsne_list(tsne_df, virus_df)
    data["marker_gene_table"] = marker_table(marker_df)

    with open(json_file, 'w') as fh:
        json.dump(data, fh)


def cluster_tsne_list(tsne_df):
    """
    tSNE_1	tSNE_2	cluster Gene_Counts
    return data list
    """
    sum_df = tsne_df.groupby(["cluster"]).agg("count").iloc[:, 0]
    percent_df = sum_df.transform(lambda x: round(x / sum(x) * 100, 2))
    res = []
    for cluster in sorted(tsne_df.cluster.unique()):
        sub_df = tsne_df[tsne_df.cluster == cluster]
        name = "cluster {cluster}({percent}%)".format(
            cluster=cluster, percent=percent_df[cluster])
        tSNE_1 = list(sub_df.tSNE_1)
        tSNE_2 = list(sub_df.tSNE_2)
        res.append({"name": name, "tSNE_1": tSNE_1, "tSNE_2": tSNE_2})
    return res


def virus_tsne_list(tsne_df, virus_df):
    """
    return data dic
    """
    tsne_df.rename(columns={"Unnamed: 0": "barcode"}, inplace=True)
    df = pd.merge(tsne_df, virus_df, on="barcode", how="left")
    df["UMI"] = df["UMI"].fillna(0)
    tSNE_1 = list(df.tSNE_1)
    tSNE_2 = list(df.tSNE_2)
    virus_UMI = list(df.UMI)
    res = {"tSNE_1": tSNE_1, "tSNE_2": tSNE_2, "virus_UMI": virus_UMI}
    return res


def marker_table(marker_df):
    """
    return html code
    """
    marker_df = marker_df.loc[:, ["cluster", "gene",
                                  "avg_logFC", "pct.1", "pct.2", "p_val_adj"]]
    marker_gene_table = marker_df.to_html(
        escape=False,
        index=False,
        table_id="marker_gene_table",
        justify="center")
    return marker_gene_table


@add_log
def seurat(sample, outdir, matrix_file):
    app = toolsdir + "/run_analysis.R"
    cmd = f"Rscript {app} --sample {sample} --outdir {outdir} --matrix_file {matrix_file}"
    os.system(cmd)


@add_log
def analysis_rna_virus(args):

    # check dir
    outdir = args.outdir
    sample = args.sample
    matrix_file = args.matrix_file
    virus_file = args.virus_file

    if not os.path.exists(outdir):
        os.system('mkdir -p %s' % (outdir))

    # run_R
    seurat(sample, outdir, matrix_file)

    # report
    tsne_df_file = glob.glob("{outdir}/*tsne_coord.tsv".format(outdir=outdir))[0]
    marker_df_file = glob.glob("{outdir}/*markers.tsv".format(outdir=outdir))[0]
    tsne_df = pd.read_csv(tsne_df_file, sep="\t")
    marker_df = pd.read_csv(marker_df_file, sep="\t")
    virus_df = pd.read_csv(virus_file, sep="\t")

    report_prepare(outdir, tsne_df, marker_df, virus_df)

    t = reporter(
        name='analysis_rna_virus',
        assay=args.assay,
        sample=args.sample,
        outdir=args.outdir + '/..')
    t.get_report()


def get_opts_analysis_rna_virus(parser, sub_program):
    if sub_program:
        s_common(parser)
        parser.add_argument('--matrix_file', help='matrix file', required=True)
        parser.add_argument(
            '--virus_file',
            help='virus UMI count file',
            required=True)
        



import numpy as np
import pandas as pd
import subprocess
import configparser
import pysam
from scipy.io import mmwrite
from scipy.sparse import csr_matrix
from celescope.tools.report import reporter
from celescope.tools.utils import *
from mutract.utils import read_CID
from celescope.tools.Analysis import Analysis


class analysis_variant(Analysis):
        
    def get_df_count_tsne(self, variant_count_file, CID_file):
        df_vc = pd.read_csv(variant_count_file, sep='\t')
        df_vc = df_vc[df_vc["alt_count"] > 0]
        df_vc_cell = df_vc.groupby('CID').agg({
            'alt_count':'count',
            'VID':list,
        })

        df_CID, df_valid = read_CID(CID_file)
        df_CID = df_CID.reset_index()
        self.tsne_df = pd.merge(self.tsne_df, df_CID, on='barcode', how='left')

        df_vc_barcode = pd.merge(df_vc_cell, df_CID, on='CID')
        df_vc_barcode_tsne = pd.merge(df_vc_barcode, self.tsne_df, on=['barcode','CID'], how='right')
        df_vc_barcode_tsne['value'] = df_vc_barcode_tsne['alt_count']
        df_vc_barcode_tsne['value'] = df_vc_barcode_tsne['value'].fillna(0)
        df_vc_barcode_tsne['value'].astype('int32')
        self.df_count_tsne = df_vc_barcode_tsne
        # out
        out_file = f'{self.outdir}/{self.sample}_count_tsne.tsv'
        df_out = df_vc_barcode_tsne[df_vc_barcode_tsne['value'] > 0]
        cols = ['CID', 'alt_count', 'VID', 'barcode', 'cluster', 'tSNE_1', 'tSNE_2', 'Gene_Counts']
        df_out = df_out[cols]
        df_out.to_csv(out_file, sep='\t')

    def get_count_tsne(self):
        def return_text(row):
            text = f'CID:{row["CID"]} <br>Variants:{str(int(row["value"]))} <br>VID:{row["VID"]}'
            return text
        tSNE_1 = list(self.df_count_tsne.tSNE_1)
        tSNE_2 = list(self.df_count_tsne.tSNE_2)
        text = list(self.df_count_tsne.apply(return_text, axis=1))
        value = list(self.df_count_tsne.value)
        title = 't-SNE plot Colored by Cell Variant Counts'
        count_tsne = {"tSNE_1": tSNE_1, "tSNE_2": tSNE_2, "text": text, 'value':value, 'title':title}
        return count_tsne

    def add_GT(self, vcf_file):
        vcf = pysam.VariantFile(vcf_file, 'r')
        out_vcf_file = f'{self.outdir}/{self.sample}_addGT.vcf'
        out_vcf = pysam.VariantFile(out_vcf_file, 'w', header=vcf.header)
        for rec in vcf:
            for sample in rec.samples:
                rec.samples[sample]["GT"] = (1,1)
                out_vcf.write(rec)
        vcf.close()
        out_vcf.close()
        return out_vcf_file


    def get_df_table(self, vcf_file, annovar_config):
        
        vcf_file = self.add_GT(vcf_file)

        df_vcf = parse_vcf(vcf_file, infos=['VID','CID'])
        df_annovar = self.annovar(vcf_file, annovar_config)
        df_vcf = pd.concat((df_vcf, df_annovar), axis=1)
        df_vcf["nCell"] = df_vcf["CID"].apply(func=lambda row:1 if isinstance(row,str) else len(row))

        out_df_vcf = f'{self.outdir}/{self.sample}_variant_table.tsv'
        df_vcf.to_csv(out_df_vcf, sep='\t', index=False)

        cols = ['VID','Chrom','Pos','Alleles','Gene','nCell','mRNA','Protein','COSMIC']
        df_vcf = df_vcf[cols]
        return df_vcf


    def run(self, variant_count_file, CID_file, vcf_file, annovar_config):
        cluster_tsne = self.get_cluster_tsne(colname='cluster')

        self.get_df_count_tsne(variant_count_file, CID_file)
        count_tsne = self.get_count_tsne()

        df_vcf = self.get_df_table(vcf_file, annovar_config)
        table_dict = Analysis.get_table(title='Variant table', id='variant_table', df_table=df_vcf)

        self.report_prepare(
            cluster_tsne=cluster_tsne,
            count_tsne=count_tsne,
            table_dict=table_dict,
        )
        self.report(stat=False)

    @add_log
    def annovar(self, vcf_file, annovar_config):

        # config
        config = configparser.ConfigParser()
        config.read(annovar_config)
        section = config['ANNOVAR']
        dir = section['dir']
        db = section['db']
        buildver = section['buildver']
        protocol = section['protocol']
        operation = section['operation']

        # convert
        input_file = f'{self.outdir}/{self.sample}.input'
        cmd = (
            f'perl {dir}/convert2annovar.pl '
            f'-format vcf4 '
            f'--includeinfo '
            f'{vcf_file} > {input_file}'
        )
        subprocess.check_call(cmd, shell=True)

        # annotate
        cmd = (
            f'perl {dir}/table_annovar.pl '
            f'{input_file} '
            f'{db} '
            f'-buildver {buildver} '
            f'-protocol {protocol} '
            f'-operation {operation} '
            f'-out {self.outdir}/{self.sample} '
            f'--otherinfo '
        )
        analysis_variant.annovar.logger.info(cmd)
        subprocess.check_call(cmd, shell=True)

        # df
        annovar_file = f'{self.outdir}/{self.sample}.{buildver}_multianno.txt'
        df_annovar = parse_annovar(annovar_file)
        return df_annovar


@add_log
def analysis_snp(args):
    step = 'analysis_snp'
    step_snp = analysis_variant(
        args.sample,
        args.outdir,
        args.assay,
        args.match_dir,
        step, 
    )
    step_snp.run(args.variant_count_file, args.CID_file, args.vcf, args.annovar_config)

def get_opts_analysis_snp(parser, sub_program):
    parser.add_argument('--annovar_config', help='annovar soft config file', required=True)
    if sub_program:
        s_common(parser)
        parser.add_argument('--match_dir', help='match_dir', required=True)
        parser.add_argument('--vcf', help='vcf file', required=True)
        parser.add_argument('--CID_file', help='CID_file', required=True)
        parser.add_argument('--variant_count_file', help='variant count file', required=True)


import unittest
import os
import pandas as pd
from collections import namedtuple
from celescope.tools.STAR import Step_mapping
from celescope.tools.utils import *
from .Chemistry import Chemistry
from celescope.tools.count import *
from celescope.tools.count import report_prepare as report_p
from celescope.tools.cutadapt import *
from celescope.tools.Multi import *

class Tests(unittest.TestCase):
    def setUp(self):
        os.chdir('/SGRNJ01/RD_dir/pipeline_test/zhouyiqi/unittest/rna')
        self.sample = 'test1'
        self.fq = './test1/02.cutadapt/test1_clean_2.fq.gz'
        self.genomeDir = '/SGRNJ/Public/Database/genome/homo_mus'
        self.thread = 4
        self.assay = 'rna'
        self.out_unmapped = False
        self.debug = True

    @unittest.skip('pass')
    def test_ribo(self):
        self.outdir = f'{self.sample}/03.STAR'
        mapping = Step_mapping(
            self.sample, 
            self.outdir, 
            self.assay, 
            self.thread,
            self.fq, 
            self.genomeDir, 
            self.out_unmapped, 
            self.debug)
        mapping.STAR_map_log = f'{self.outdir}/{self.sample}_Log.final.out'  
        mapping.STAR_map_log = f'{self.outdir}/{self.sample}_Log.final.out'
        mapping.picard_region_log = f'{self.outdir}/{self.sample}_region.log'
        mapping.ribo()
        mapping.format_stat()
        mapping.report()

    @unittest.skip('pass')
    def test_chemistry(self):
        fq1s = [
            '/SGRNJ/DATA_PROJ/2004/20201102_9/MG_201026_1-R2010283_L3_1.fq.gz',
            '/SGRNJ01/RD_dir/pipeline_test/zhouyiqi/unittest/test_data/smk/R2009217_L4_1.fq.gz',
            '/SGRNJ/DATA_PROJ/2003/20200727_6/R2007185_L3_1.fq.gz',
            '/SGRNJ/DATA_PROJ/2004/20201029/MG_201023_1-R2010226_L3_1.fq.gz',
        ]
        results = []
        for fq1 in fq1s:
            ch = Chemistry(fq1)
            results.append(ch.check_chemistry())
        print(results)
        assert results == ['scopeV2.1.1', 'scopeV2.1.1', 'scopeV2.0.1', 'scopeV2.2.1']
        fq = '/SGRNJ03/DATA03/2004/20201207_10/R2011312-PCRD-201203-1_combined_R1.fastq.gz'
        ch = Chemistry(fq)
        print(ch.check_chemistry())

    @unittest.skip('pass')
    def test_gtf(self):
        '''
        gtf_file = '/SGRNJ/Database/script/genome/hs/gtf/Homo_sapiens.GRCh38.99.gtf'
        id_name = gene_convert(gtf_file)
        print(f"ENSG00000001629: {id_name['ENSG00000001629']}")

        gtf_file = '/SGRNJ01/RD_dir/pipeline_test/litao/genomes/Cricetulus_griseus/Cricetulus_griseus_crigri.CriGri_1.0.101.gtf'
        id_name = gene_convert(gtf_file)
        print(id_name)
        '''

        gtf_file = '/Public/Database/genome/Sus_scrofa/ncbi/GCF_000003025.6_Sscrofa11.1_genomic_new.gtf'
        id_name = gene_convert(gtf_file)
        print(id_name['tRNA-Asp'])

    @unittest.skip('pass')
    def test_rescue(self):
        tools_dir = os.path.dirname(__file__)
        os.chdir('/SGRNJ01/RD_dir/pipeline_test/zhouyiqi/unittest/rna')
        outdir = 'test2/05.count'
        sample = 'test2'
        matrix_dir = f'{outdir}/{sample}_all_matrix_10X'
        app = f'{tools_dir}/rescue.R'
        cmd = (
            f'Rscript {app} '
            f'--outdir {outdir} '
            f'--sample {sample} '
            f'--matrix_dir {matrix_dir} '
            f'--threshold 14 '
        )
        print(cmd)
        subprocess.check_call(cmd, shell=True)
    
    @unittest.skip('pass')
    def test_get_fq(self):
        library_id = 'R2011312'
        library_path = '/SGRNJ03/DATA03/2004/20201122_14/'
        fq1, fq2 = get_fq(library_id, library_path)
        print(fq1, fq2)

    @unittest.skip('pass')
    def test_matrix_10X(self):
        match_dir = '/SGRNJ01/RD_dir/pipeline_test/zhouyiqi/unittest/rna/test2/'
        validated_barcodes, _ncell = read_barcode_file(match_dir)    
        os.chdir('/SGRNJ01/RD_dir/pipeline_test/zhouyiqi/unittest/rna/')
        df = pd.read_csv('/SGRNJ01/RD_dir/pipeline_test/zhouyiqi/unittest/rna/test2/05.count/test2_count_detail.txt.gz',sep='\t')
        outdir = 'test2/05.count'
        sample = 'test2'
        gtf_file = '/SGRNJ/Public/Database/genome/homo_sapiens/ensembl_92/Homo_sapiens.GRCh38.92.chr.gtf'
        matrix_10X(df, outdir, sample, gtf_file, dir_name='matrix_10X_new', validated_barcodes=validated_barcodes)

    @unittest.skip('pass')
    def test_call_cells(self):
        os.chdir('/SGRNJ03/PROJ03/PROJ_20.SC/PN20122407_SCOPEv2/temp/CK2101_origin/')
        sample = 'CK2101'
        outdir = f'{sample}/05.count'
        df = pd.read_csv(f'{outdir}/{sample}_count_detail.txt.gz', sep='\t')
        outdir = f'{sample}/05.count'
        cell_calling_method = 'cellranger3'
        force_cell_num = None
        expected_cell_num = 3000
        all_matrix_10X_dir = f'{outdir}/{sample}_all_matrix'

        # df_sum
        df_sum = get_df_sum(df)

        # call cells
        cell_bc, threshold = cell_calling(
            cell_calling_method, force_cell_num, expected_cell_num, all_matrix_10X_dir, df_sum, outdir, sample)

        # plot
        cell_num = len(cell_bc)
        print(f'cell_num:{cell_num}')
        #plot_barcode_UMI(df_sum, threshold, expected_cell_num, cell_num, outdir, sample, cell_calling_method,col='UMI')


    @unittest.skip('pass')
    def test_read_adapter_fasta(self):
        os.chdir('/SGRNJ01/RD_dir/pipeline_test/zhouyiqi/unittest/rna')
        adapter_fasta = './adapter.fasta'
        adapter_args = read_adapter_fasta(adapter_fasta)
        assert adapter_args == ['a1=ATCG','a2=TGCAA']

    @unittest.skip('pass')
    def test_chemistry_jiace(self):
        fq = '/SGRNJ03/DATA03/2004/20201224_18/R20067441-PCRD-201207-1-R2012093_combined_R1.fastq.gz\
,/SGRNJ03/DATA03/2004/20201228_6/R2012093-PCRD-201207-1_combined_R1.fastq.gz'
        ch = Chemistry(fq)
        print(ch.check_chemistry())

    '''
    @unittest.skip('pass')
    def test_report_prepare(self):
        outdir = "/SGRNJ03/randd/P19112803_SCOPEv1/test1/NJXK01_1/05.count"
        count_file = f"{outdir}/NJXK01_1_counts.txt"
        downsample_file = f"{outdir}/NJXK01_1_downsample.txt"
        report_prepare(count_file, downsample_file, outdir)
    '''

    @unittest.skip('pass')
    def test_Multi(self):
        os.chdir('/SGRNJ01/RD_dir/pipeline_test/zhouyiqi/unittest/rna/rebuild')
        multi = Multi('rna')

        sys.argv = ['celescope', '--mapfile', 'test.mapfile', '--not_gzip']
        '''
        multi.parse_args()
        print(multi.parse_step_args('sample'))
        multi.prepare()
        multi.sample('test1')
        multi.barcode('test1')
        print(multi.sjm_cmd)
        '''

        multi.run()
        print(multi.sjm_cmd)

    @unittest.skip('pass')
    @add_mem
    def test_downsample(self):
        assay = 'rna'
        step = 'count'
        outdir = '/SGRNJ01/RD_dir/pipeline_test/zhouyiqi/unittest/rna/rebuild/test1/05.count/'
        sample = 'test1'
        count_detail_file = f'{outdir}/{sample}_count_detail.txt.gz'
        cell_bc, _ = read_barcode_file(f"{outdir}/../")
        df = pd.read_table(count_detail_file, header=0)
        df_cell = df.loc[df["Barcode"].isin(cell_bc), :]

        json_file = f'{outdir}/data.json'
        reporter = Reporter(assay, 'count', sample, outdir, json_file)
        downsample_file = "/SGRNJ01/RD_dir/pipeline_test/zhouyiqi/unittest/rna/rebuild/test1/05.count/new.downsample"
        Saturation, res_dict = downsample(df, cell_bc, downsample_file)
        print(res_dict)
        reporter.add_data_item(downsample_summary=res_dict)

        reporter.dump_json()
    
    @unittest.skip('pass')
    @add_mem
    def test_downsample_large(self):
        count_detail_file = '/SGRNJ03/randd/P19112803_SCOPEv1/test1/NJXK01_1/old.count/NJXK01_1_count_detail.txt.gz'
        cell_bc, _ = read_barcode_file("/SGRNJ03/randd/P19112803_SCOPEv1/test1/NJXK01_1/")
        df = pd.read_table(count_detail_file, header=0)
        #df_cell = df.loc[df["Barcode"].isin(cell_bc), :]

        downsample_file = "/SGRNJ03/randd/P19112803_SCOPEv1/test1/NJXK01_1/05.count/test.downsample"
        downsample(df, cell_bc, downsample_file)
    
    @unittest.skip('pass')
    @add_mem
    def test_downsample_real1(self):
        count_detail_file = '/SGRNJ03/PROJ03/PROJ_20.SC/PN20122407_SCOPEv2/temp/CK2101_origin/CK2101/05.count/CK2101_count_detail.txt.gz'
        cell_bc, _ = read_barcode_file("/SGRNJ03/PROJ03/PROJ_20.SC/PN20122407_SCOPEv2/temp/CK2101_origin/CK2101/")
        df = pd.read_table(count_detail_file, header=0)
        #df_cell = df.loc[df["Barcode"].isin(cell_bc), :]

        downsample_file = "/SGRNJ01/RD_dir/pipeline_test/zhouyiqi/downsample/CK2101_origin.downsample.txt"
        downsample(df, cell_bc, downsample_file)

    @unittest.skip('pass')
    @add_mem
    def test_downsample_large1(self):
        assay = 'rna'
        step = 'count'
        outdir = '/SGRNJ03/PROJ03/PROJ_20.SC/PN20122407_SCOPEv2/temp/CK2101_origin/CK2101/05.count/'
        sample = 'CK2101'
        count_detail_file = f'{outdir}/{sample}_count_detail.txt.gz'
        cell_bc, _ = read_barcode_file(f"{outdir}/../")
        df = pd.read_table(count_detail_file, header=0)
        df_cell = df.loc[df["Barcode"].isin(cell_bc), :]

        json_file = f'/SGRNJ01/RD_dir/pipeline_test/zhouyiqi/downsample/CK2101_origin.json'
        reporter = Reporter(assay, 'count', sample, outdir, json_file)
        downsample_file = "/SGRNJ01/RD_dir/pipeline_test/zhouyiqi/downsample/CK2101_origin.downsample.txt"
        Saturation, res_dict = downsample(df, cell_bc, downsample_file)
        print(res_dict)
        reporter.add_data_item(downsample_summary=res_dict)

        reporter.dump_json()

    @unittest.skip('pass')
    @add_mem
    def test_downsample_large2(self):
        assay = 'rna'
        step = 'count'
        outdir = '/SGRNJ03/PROJ03/PROJ_20.SC/PN20122407_SCOPEv2/temp/CK816_1-1/CK816_1-1/05.count/'
        sample = 'CK816_1-1'
        count_detail_file = f'{outdir}/{sample}_count_detail.txt.gz'
        cell_bc, _ = read_barcode_file(f"{outdir}/../")
        df = pd.read_table(count_detail_file, header=0)
        df_cell = df.loc[df["Barcode"].isin(cell_bc), :]

        json_file = f'/SGRNJ01/RD_dir/pipeline_test/zhouyiqi/downsample/CK816_1-1.json'
        reporter = Reporter(assay, 'count', sample, outdir, json_file)
        downsample_file = "/SGRNJ01/RD_dir/pipeline_test/zhouyiqi/downsample/CK816_1-1.downsample.txt"
        Saturation, res_dict = downsample(df, cell_bc, downsample_file)
        print(res_dict)
        reporter.add_data_item(downsample_summary=res_dict)

        reporter.dump_json()

    @unittest.skip('pass')
    @add_mem
    def test_report_prepare(self):
        os.chdir("/SGRNJ03/randd/P19112803_SCOPEv1/test1/")
        marked_counts_file = 'NJXK01_1/05.count/NJXK01_1_counts.txt'
        downsample_file = 'NJXK01_1/05.count/NJXK01_1_downsample.txt'
        outdir = 'NJXK01_1/05.count/'
        report_p(marked_counts_file, downsample_file, outdir)

    @unittest.skip('pass')
    def test_report(self):
        os.chdir("/SGRNJ03/randd/P19112803_SCOPEv1/test1/")
        assay = 'rna'
        sample = 'NJXK01_1'
        outdir = 'NJXK01_1/05.count/'
        t = reporter(assay=assay,
                 name='count', sample=sample,
                 stat_file=outdir + '/stat.txt',
                 outdir=outdir + '/..')
        t.get_report()

    @unittest.skip('pass')
    def test_metrics(self):
        assay = 'rna'
        sample = 'test1'
        __STEPS__ = [
            'sample',
            'barcode',
            'cutadapt',
            'STAR',
            "featureCounts",
            "count",
            'analysis',
        ]
        for index, step in enumerate(__STEPS__):
            print(index, step)
            outdir = f'/SGRNJ01/RD_dir/pipeline_test/zhouyiqi/unittest/rna/rebuild/test1/0{index}.{step}'
            report = Reporter(
                assay,
                step,
                sample,
                outdir,
            )
            report.stat_to_json()
            report.dump_json()

    def test_sorted_dumb_consensus(self):
        os.chdir('/SGRNJ01/RD_dir/pipeline_test/zhouyiqi/unittest/vdj/rebuild/TCR_20201221_consensus/03.consensus')
        fastq = 'TCR_20201221_consensus_sorted.fq.tmp'
        sorted_dumb_consensus(fastq, outfile = 'update_consensus.fq.gz', threshold=0.5)


import unittest
import os
import pandas as pd
from celescope.hla.mapping_hla import split_bam, hla_typing, summary, read_index
from celescope.tools.utils import *


class testHLA(unittest.TestCase):
    def setUp(self):
        os.chdir('/SGRNJ01/RD_dir/pipeline_test/zhouyiqi/HLA/test_0908/')
        self.sample = 'pT_HLA_0812'
        self.out_bam = './/pT_HLA_0812/03.mapping_hla/bam/pT_HLA_0812.bam'
        self.match_dir = '/SGRNJ01/RD_dir/pipeline_test/zhouyiqi/HLA/test_0908/match_dir/'
        self.barcodes, _ncell = read_barcode_file(self.match_dir)
        self.thread = 6
        self.mapping_outdir = f'{self.sample}/03.mapping_hla/'
        self.index_file = f'{self.mapping_outdir}/{self.sample}_cell_index.tsv'

    @unittest.skip('pass')
    def test_split_bam(self):
        outdir = self.mapping_outdir
        split_bam(self.out_bam, self.barcodes, outdir, self.sample)

    @unittest.skip('pass')
    def test_hla_typing(self):
        hla_typing(self.index_file, self.mapping_outdir, self.thread)

    @unittest.skip('pass')
    def test_read_index(self):
        read_index(self.index_file)

    #@unittest.skip('pass')
    def test_summary(self):
        summary(self.index_file, self.mapping_outdir, self.sample)


if __name__ == '__main__':
    unittest.main()



import pysam
import numpy as np
import subprocess
from xopen import xopen
from collections import defaultdict
from itertools import groupby

from celescope.tools.Step import Step, s_common
import celescope.tools.utils as utils


@utils.add_log
def sort_fastq(fq, fq_tmp_file, outdir):
    tmp_dir = f'{outdir}/tmp'
    cmd = (
        f'mkdir {tmp_dir};'
        f'less {fq} | paste - - - - | sort -T {tmp_dir} -k1,1 -t " " | tr "\t" "\n" > {fq_tmp_file};'
    )
    subprocess.check_call(cmd, shell=True)


@utils.add_log
def sorted_dumb_consensus(fq, outfile, threshold):
    '''
    consensus read in name-sorted fastq
    output (barcode,umi) consensus fastq
    '''
    read_list = []
    n_umi = 0
    total_ambiguous_base_n = 0
    length_list = []
    out_h = xopen(outfile, 'w')

    def keyfunc(read):
        attr = read.name.split('_')
        return (attr[0], attr[1])

    with pysam.FastxFile(fq) as fh:
        for (barcode, umi), g in groupby(fh, key=keyfunc):
            read_list = []
            for read in g:
                read_list.append([read.sequence, read.quality])
            consensus_seq, consensus_qual, ambiguous_base_n, con_len = dumb_consensus(
                read_list, threshold=threshold, ambiguous="N")
            n_umi += 1
            prefix = "_".join([barcode, umi])
            read_name = f'{prefix}_{n_umi}'
            out_h.write(utils.fastq_line(read_name, consensus_seq, consensus_qual))
            if n_umi % 10000 == 0:
                sorted_dumb_consensus.logger.info(f'{n_umi} UMI done.')
            total_ambiguous_base_n += ambiguous_base_n
            length_list.append(con_len)
    
    out_h.close()
    return n_umi, total_ambiguous_base_n, length_list


@utils.add_log
def wrap_consensus(fq, outdir, sample, threshold):
    fq_tmp_file = f'{outdir}/{sample}_sorted.fq.tmp'
    sort_fastq(fq, fq_tmp_file, outdir)
    outfile = f'{outdir}/{sample}_consensus.fq'
    n, total_ambiguous_base_n, length_list = sorted_dumb_consensus(
        fq=fq_tmp_file, outfile=outfile, threshold=threshold)
    return outfile, n, total_ambiguous_base_n, length_list


def dumb_consensus(read_list, threshold=0.5, ambiguous='N', default_qual='F'):
    '''
    This is similar to biopython dumb_consensus.
    It will just go through the sequence residue by residue and count up the number of each type
    of residue (ie. A or G or T or C for DNA) in all sequences in the
    alignment. If the percentage of the most common residue type is
    greater then the passed threshold, then we will add that residue type,
    otherwise an ambiguous character will be added.
    elements of read_list: [entry.sequence,entry.quality]
    '''

    con_len = get_read_length(read_list, threshold=threshold)
    consensus_seq = ""
    consensus_qual = ""
    ambiguous_base_n = 0
    for n in range(con_len):
        atom_dict = defaultdict(int)
        quality_dict = defaultdict(int)
        num_atoms = 0
        for read in read_list:
            # make sure we haven't run past the end of any sequences
            # if they are of different lengths
            sequence = read[0]
            quality = read[1]
            if n < len(sequence):
                atom = sequence[n]
                atom_dict[atom] += 1
                num_atoms = num_atoms + 1

                base_qual = quality[n]
                quality_dict[base_qual] += 1

        consensus_atom = ambiguous
        for atom in atom_dict:
            if atom_dict[atom] >= num_atoms * threshold:
                consensus_atom = atom
                break
        if consensus_atom == ambiguous:
            ambiguous_base_n += 1
        consensus_seq += consensus_atom

        max_freq_qual = 0
        consensus_base_qual = default_qual
        for base_qual in quality_dict:
            if quality_dict[base_qual] > max_freq_qual:
                max_freq_qual = quality_dict[base_qual]
                consensus_base_qual = base_qual

        consensus_qual += consensus_base_qual
    return consensus_seq, consensus_qual, ambiguous_base_n, con_len


def get_read_length(read_list, threshold=0.5):
    '''
    compute read_length from read_list. 
    length = max length with read fraction >= threshold
    elements of read_list: [entry.sequence,entry.quality]
    '''
    
    n_read = len(read_list)
    length_dict = defaultdict(int)
    for read in read_list:
        length = len(read[0])
        length_dict[length] += 1
    for length in length_dict:
        length_dict[length] = length_dict[length] / n_read

    fraction = 0
    for length in sorted(length_dict.keys(),reverse=True):
        fraction += length_dict[length]
        if fraction >= threshold:
            return length

@utils.add_log
def consensus(args):

    step_name = "consensus"
    step = Step(args, step_name)

    if args.not_consensus:
        consensus.logger.warning("not_consensus!")
        return
    sample = args.sample
    outdir = args.outdir
    assay = args.assay
    fq = args.fq
    threshold = float(args.threshold)

    outfile, n, total_ambiguous_base_n, length_list = wrap_consensus(fq, outdir, sample, threshold)

    # metrics
    metrics = {}
    metrics["UMI Counts"] = n
    metrics["Mean UMI Length"] = np.mean(length_list)
    metrics["Ambiguous Base Counts"] = total_ambiguous_base_n    
    utils.format_metrics(metrics)

    ratios = {}
    ratios["Ambiguous Base Counts Ratio"] = total_ambiguous_base_n / sum(length_list)
    utils.format_ratios(ratios)

    # stat file
    stat_file = f'{outdir}/stat.txt'
    with open(stat_file, 'w') as stat_h:
        stat_str = (
            f'UMI Counts: {metrics["UMI Counts"]}\n'
            f'Mean UMI Length: {metrics["Mean UMI Length"]}\n'
            f'Ambiguous Base Counts: {metrics["Ambiguous Base Counts"]}({ratios["Ambiguous Base Counts Ratio"]}%)\n'
        )
        stat_h.write(stat_str)

    step.clean_up()



def get_opts_consensus(parser, sub_program):
    parser.add_argument("--threshold", help='valid base threshold', default=0.5)
    parser.add_argument("--not_consensus", help="input fastq is not consensus", action='store_true')
    if sub_program:
        s_common(parser)
        parser.add_argument("--fq", required=True)


import argparse
from .Count_tag import Count_tag
from celescope.tools.Step import Step, s_common


def get_opts_count_tag(parser, sub_program):
    parser.add_argument(
        "--UMI_min",
        help="cells have tag_UMI>=UMI_min are considered as valid cell",
        default="auto")
    parser.add_argument("--dim", help="tag dimension", default=1)
    parser.add_argument(
        "--SNR_min",
        help="minimum signal to noise ratio",
        default="auto")
    parser.add_argument(
        "--combine_cluster",
        help="conbine cluster tsv file",
        default=None)
    parser.add_argument("--coefficient", "-c", help="SNR coefficient", default=0.1)
    if sub_program:
        parser = s_common(parser)
        parser.add_argument("--read_count_file", help="tag read count file")
        parser.add_argument("--match_dir", help="matched scRNA-Seq CeleScope directory path", required=True)


def count_tag(args):

    step_name = "count_tag"
    step = Step(args, step_name)

    count_tag_object = Count_tag(
        args.sample,
        args.outdir,
        args.assay,
        args.read_count_file,
        args.match_dir,
        args.UMI_min,
        args.SNR_min,
        args.combine_cluster,
        args.dim,
        args.coefficient
    )
    count_tag_object.run()
    count_tag_object.stat()

    step.clean_up()


import os
import sys
import json
import functools
from collections import defaultdict
from itertools import groupby
import numpy as np
import pandas as pd
from scipy.io import mmwrite
from scipy.sparse import csr_matrix
import pysam
from celescope.tools.utils import *
from celescope.tools.report import reporter


def report_prepare(count_file, downsample_file, outdir):

    json_file = outdir + '/.data.json'
    if not os.path.exists(json_file):
        data = {}
    else:
        fh = open(json_file)
        data = json.load(fh)
        fh.close()

    df0 = pd.read_table(downsample_file, header=0)
    data['percentile'] = df0['percent'].tolist()
    data['MedianGeneNum'] = df0['median_geneNum'].tolist()
    data['Saturation'] = df0['saturation'].tolist()

    #data['count' + '_summary'] = df0.T.values.tolist()

    df = pd.read_table(count_file, header=0)
    df = df.sort_values('UMI', ascending=False)
    data['CB_num'] = df[df['mark'] == 'CB'].shape[0]
    data['Cells'] = list(df.loc[df['mark'] == 'CB', 'UMI'])

    data['UB_num'] = df[df['mark'] == 'UB'].shape[0]
    data['Background'] = list(df.loc[df['mark'] == 'UB', 'UMI'])

    data['umi_summary'] = True

    with open(json_file, 'w') as fh:
        json.dump(data, fh)


@add_log
def barcode_filter_with_magnitude(
        df, expected_cell_num, plot='magnitude.pdf', col='UMI', percent=0.1):
    # col can be readcount or UMI
    # determine validated barcodes

    df = df.sort_values(col, ascending=False)
    if expected_cell_num == "auto":
        idx = int(3000 * 0.01)
        barcode_number = df.shape[0]
        idx = int(min(barcode_number, idx))
        if idx == 0:
            sys.exit("cell number equals zero!")
        # calculate read counts threshold
        threshold = int(df.iloc[idx - 1, df.columns == col] * 0.1)
        threshold = max(1, threshold)
    else:
        expected_cell_num = int(expected_cell_num)
        cell_range = int(expected_cell_num * 0.1)
        cell_low = expected_cell_num - cell_range
        cell_high = expected_cell_num + cell_range

        df_barcode_count = df.groupby(
            ['UMI']).size().reset_index(
            name='barcode_counts')
        sorted_df = df_barcode_count.sort_values("UMI", ascending=False)
        sorted_df["barcode_cumsum"] = sorted_df["barcode_counts"].cumsum()
        for i in range(sorted_df.shape[0]):
            if sorted_df.iloc[i, :]["barcode_cumsum"] >= cell_low:
                index_low = i - 1
                break
        for i in range(sorted_df.shape[0]):
            if sorted_df.iloc[i, :]["barcode_cumsum"] >= cell_high:
                index_high = i
                break
        df_sub = sorted_df.iloc[index_low:index_high + 1, :]
        threshold = df_sub.iloc[np.argmax(
            np.diff(df_sub["barcode_cumsum"])), :]["UMI"]
        barcode_filter_with_magnitude.logger.info(
            "UMI threshold: " + str(threshold))
    validated_barcodes = df[df[col] >= threshold].index

    import matplotlib
    matplotlib.use('Agg')
    import matplotlib.pyplot as plt
    fig = plt.figure()
    plt.plot(df[col])
    plt.hlines(threshold, 0, len(validated_barcodes), linestyle='dashed')
    plt.vlines(len(validated_barcodes), 0, threshold, linestyle='dashed')
    plt.title('expected cell num: %s\n%s threshold: %s\ncell num: %s' %
              (expected_cell_num, col, threshold, len(validated_barcodes)))
    plt.loglog()
    plt.savefig(plot)

    return (validated_barcodes, threshold, len(validated_barcodes))


def hd(x, y):
    return len([i for i in range(len(x)) if x[i] != y[i]])


def correct_umi(fh1, barcode, gene_umi_dict, percent=0.1):
    res_dict = defaultdict()

    for geneID in gene_umi_dict:
        _dict = gene_umi_dict[geneID]
        umi_arr = sorted(
            _dict.keys(), key=lambda x: (_dict[x], x), reverse=True)
        while True:
            # break when only one barcode or umi_low/umi_high great than 0.1
            if len(umi_arr) == 1:
                break
            umi_low = umi_arr.pop()

            for u in umi_arr:
                if float(_dict[umi_low]) / _dict[u] > percent:
                    break
                if hd(umi_low, u) == 1:
                    _dict[u] += _dict[umi_low]
                    del (_dict[umi_low])
                    break
        res_dict[geneID] = _dict
    return res_dict


@add_log
def bam2table(bam, detail_file, id_name):
    # 提取bam中相同barcode的reads，统计比对到基因的reads信息
    # probe
    probe_gene_count_dict = genDict(dim=4, valType=int)

    samfile = pysam.AlignmentFile(bam, "rb")
    with open(detail_file, 'w') as fh1:
        fh1.write('\t'.join(['Barcode', 'geneID', 'UMI', 'count']) + '\n')

        # pysam.libcalignedsegment.AlignedSegment
        # AAACAGGCCAGCGTTAACACGACC_CCTAACGT_A00129:340:HHH72DSXX:2:1353:23276:30843
        # 获取read的barcode
        def keyfunc(x): return x.query_name.split('_', 1)[0]

        for _, g in groupby(samfile, keyfunc):
            gene_umi_dict = defaultdict(lambda: defaultdict(int))
            for seg in g:
                (barcode, umi, probe) = seg.query_name.split('_')[:3]
                if probe != 'None':
                    probe_gene_count_dict[probe]['total'][barcode][umi] += 1
                    if seg.has_tag('XT'):
                        geneID = seg.get_tag('XT')
                        geneName = id_name[geneID]
                        probe_gene_count_dict[probe][geneName][barcode][umi] += 1
                    else:
                        probe_gene_count_dict[probe]['None'][barcode][umi] += 1
                if not seg.has_tag('XT'):
                    continue
                geneID = seg.get_tag('XT')
                gene_umi_dict[geneID][umi] += 1
            res_dict = correct_umi(fh1, barcode, gene_umi_dict)

            # output
            for geneID in res_dict:
                for umi in res_dict[geneID]:
                    fh1.write('%s\t%s\t%s\t%s\n' % (barcode, geneID, umi,
                                                    res_dict[geneID][umi]))

    # out probe
    row_list = []
    for probe in probe_gene_count_dict:
        for geneName in probe_gene_count_dict[probe]:
            barcode_count = len(probe_gene_count_dict[probe][geneName])
            umi_count = 0
            read_count = 0
            for barcode in probe_gene_count_dict[probe][geneName]:
                for umi in probe_gene_count_dict[probe][geneName][barcode]:
                    umi_count += len( probe_gene_count_dict[probe][geneName][barcode])
                    read_count += probe_gene_count_dict[probe][geneName][barcode][umi]
            row_list.append({
                'probe': probe,
                'gene': geneName,
                'barcode_count': barcode_count,
                'read_count': read_count,
                'UMI_count': umi_count
            })

    df_probe = pd.DataFrame(row_list,
        columns=['probe', 'gene', 'barcode_count', 'read_count', 'UMI_count'])
    df_probe = df_probe.groupby(['probe']).apply(
        lambda x: x.sort_values('UMI_count', ascending=False)
    )
    return df_probe


@add_log
def call_cells(df, expected_num, pdf, marked_counts_file):
    def num_gt2(x):
        return pd.Series.sum(x[x > 1])

    df_sum = df.groupby('Barcode').agg({
        'count': ['sum', num_gt2],
        'UMI': 'count',
        'geneID': 'nunique'
    })
    df_sum.columns = ['readcount', 'UMI2', 'UMI', 'geneID']
    (validated_barcodes, threshold, cell_num) = barcode_filter_with_magnitude(
        df_sum, col='UMI', plot=pdf, expected_cell_num=expected_num)
    df_sum.loc[:, 'mark'] = 'UB'
    df_sum.loc[df_sum.index.isin(validated_barcodes), 'mark'] = 'CB'
    df_sum.to_csv(marked_counts_file, sep='\t')
    CB_describe = df_sum.loc[df_sum['mark'] == 'CB', :].describe()
    #Saturation = (1 - tmp.loc[tmp['UMI'] == 1, :].shape[0] / total)*100
    #Saturation = (df_sum['UMI2'].sum() + 0.0) / df_sum['UMI'].sum() * 100
    del df_sum

    return validated_barcodes, threshold, cell_num, CB_describe


@add_log
def expression_matrix(
        df, validated_barcodes,
        outdir, sample, id_name,
        sc_cell_barcodes, sc_cell_number):

    matrix_10X_dir = f"{outdir}/{sample}_matrix_10X/"
    matrix_table_file = f"{outdir}/{sample}_matrix.tsv.gz"
    if not os.path.exists(matrix_10X_dir):
        os.mkdir(matrix_10X_dir)

    df.loc[:, 'mark'] = 'UB'
    df.loc[df['Barcode'].isin(validated_barcodes), 'mark'] = 'CB'

    CB_total_Genes = df.loc[df['mark'] == 'CB', 'geneID'].nunique()
    CB_reads_count = df.loc[df['mark'] == 'CB', 'count'].sum()
    reads_mapped_to_transcriptome = df['count'].sum()

    table = df.loc[df['mark'] == 'CB', :].pivot_table(
        index='geneID', columns='Barcode', values='UMI',
        aggfunc=len).fillna(0).astype(int)

    id = table.index.to_series()
    name = id.apply(lambda x: id_name[x])
    genes = pd.concat([id, name], axis=1)
    genes.columns = ['gene_id', 'gene_name']

    # write 10X matrix
    table.columns.to_series().to_csv(
        f'{matrix_10X_dir}/barcodes.tsv', index=False, sep='\t')
    genes.to_csv(
        f'{matrix_10X_dir}/genes.tsv', index=False, header=False, sep='\t')
    mmwrite(f'{matrix_10X_dir}/matrix', csr_matrix(table))

    # convert id to name; write table matrix
    table.index = name
    table.index.name = ""
    table.to_csv(
        matrix_table_file,
        sep="\t",
        compression='gzip')
    
    # match median and mean
    match_cell = set(validated_barcodes).intersection(sc_cell_barcodes)
    n_match_cell = len(match_cell)
    match_percent = round(n_match_cell / sc_cell_number * 100, 2)
    match_cell_str = f'{format_number(n_match_cell)}({match_percent}%)'
    
    df_match = table.loc[:, match_cell]
    df_match.to_csv(
        f"{outdir}/{sample}_match_matrix.tsv.gz",
        sep="\t",
        compression='gzip')

    match_UMI_median = df_match.apply(np.median, axis=1)
    match_UMI_mean = df_match.apply(np.mean, axis=1)
    match_UMI_median_over_zero = df_match.apply(lambda row: np.median(row[row > 0]), axis=1)
    match_UMI_mean_over_zero = df_match.apply(lambda row: np.mean(row[row > 0]), axis=1)

    df_count = pd.DataFrame({
        "UMI_median": match_UMI_median,
        'UMI_mean': match_UMI_mean,
        "UMI_median_over_zero": match_UMI_median_over_zero,
        "UMI_mean_over_zero": match_UMI_mean_over_zero,

    })
    df_count = df_count.sort_values(['UMI_median', 'UMI_mean'], ascending=False)
    df_count.to_csv(f'{outdir}/{sample}_match_UMI_count.tsv', sep='\t')

    # median count
    df_match_cell_UMI_count = df_match.apply(sum, axis=0)
    match_cell_UMI_median = int(np.median(df_match_cell_UMI_count))

    return(CB_total_Genes, CB_reads_count, reads_mapped_to_transcriptome, match_cell_str, match_cell_UMI_median)


def get_summary(df, sample, Saturation, CB_describe, CB_total_Genes,
                CB_reads_count, reads_mapped_to_transcriptome,
                match_cell_str, match_UMI_median,
                stat_file, outdir):

    # total read
    json_file = outdir + '.data.json'
    fh = open(json_file)
    data = json.load(fh)
    str_number = data['barcode_summary'][1][1].split("(")[0]
    valid_read_number = int(str_number.replace(",", ""))
    index = [
        'Estimated Number of Cells',
        'Fraction Reads in Cells',
        'Mean Reads per Cell',
        'Median UMI per Cell',
        'Total Genes',
        'Median Genes per Cell',
        'Saturation',
        'Matched Cells',
        'Median UMI per Matched Cells',
    ]
    summary = pd.Series([0] * len(index), index=index)

    # 细胞数
    summary['Estimated Number of Cells'] = int(
        CB_describe.loc['count', 'readcount'])
    summary['Fraction Reads in Cells'] = '%.2f%%' % (float(
        CB_reads_count) / reads_mapped_to_transcriptome * 100)
    summary['Mean Reads per Cell'] = int(
        valid_read_number /
        summary['Estimated Number of Cells'])
    summary['Median UMI per Cell'] = int(CB_describe.loc['50%', 'UMI'])
    summary['Total Genes'] = int(CB_total_Genes)
    summary['Median Genes per Cell'] = int(CB_describe.loc['50%', 'geneID'])
    summary['Saturation'] = '%.2f%%' % (Saturation)
    summary['Matched Cells'] = match_cell_str
    summary['Median UMI per Matched Cells'] = match_UMI_median
    # 测序饱和度，认定为细胞中的reads中UMI>2的reads比例
    need_format = [
        'Estimated Number of Cells',
        'Mean Reads per Cell',
        'Median UMI per Cell',
        'Total Genes',
        'Median Genes per Cell',
        'Median UMI per Matched Cells',
    ]
    for item in need_format:
        summary[item] = format_number(summary[item])
    summary.to_csv(stat_file, header=False, sep=':')


def sample(p, df, bc):
    tmp = df.sample(frac=p)
    format_str = "%.2f\t%.2f\t%.2f\n"
    tmp.reset_index(drop=True, inplace=True)
    tmp = tmp.loc[tmp['Barcode'].isin(bc), :]
    geneNum_median = tmp.groupby('Barcode').agg({
        'geneID': 'nunique'
    })['geneID'].median()
    tmp = tmp.groupby(['Barcode', 'geneID', 'UMI']).agg({'UMI': 'count'})
    total = float(tmp['UMI'].count())
    saturation = (1 - tmp.loc[tmp['UMI'] == 1, :].shape[0] / total) * 100
    return format_str % (p, geneNum_median, saturation), saturation


@add_log
def downsample(detail_file, validated_barcodes, downsample_file):
    df = pd.read_table(detail_file, index_col=[0, 1, 2])
    df = df.index.repeat(df['count']).to_frame()
    format_str = "%.2f\t%.2f\t%.2f\n"
    with open(downsample_file, 'w') as fh:
        fh.write('percent\tmedian_geneNum\tsaturation\n')
        fh.write(format_str % (0, 0, 0))

        saturation = 0
        for p in np.arange(0.1, 1.1, 0.1):
            r, s = sample(p=p, df=df, bc=validated_barcodes)
            fh.write(r)
            saturation = s
    return saturation


@add_log
def count_capture_rna(args):

    # check
    _refFlat, gtf = glob_genomeDir(args.genomeDir)
    id_name = gene_convert(gtf)

    # 检查和创建输出目录
    if not os.path.exists(args.outdir):
        os.system('mkdir -p %s' % (args.outdir))

    # umi纠错，输出Barcode geneID  UMI     count为表头的表格
    count_detail_file = args.outdir + '/' + args.sample + '_count_detail.txt'
    df_probe = bam2table(args.bam, count_detail_file, id_name)
    df_probe.to_csv(f'{args.outdir}/{args.sample}_probe_gene_count.tsv', sep='\t', index=False)

    df = pd.read_table(count_detail_file, header=0)

    # call cells
    pdf = args.outdir + '/barcode_filter_magnitude.pdf'
    marked_counts_file = args.outdir + '/' + args.sample + '_counts.txt'
    (validated_barcodes, threshold, cell_num, CB_describe) = call_cells(
        df, args.cells, pdf, marked_counts_file)
    
    # match barcode
    sc_cell_barcodes, sc_cell_number = read_barcode_file(args.match_dir)

    # 输出matrix
    (
        CB_total_Genes,
        CB_reads_count,
        reads_mapped_to_transcriptome,
        match_cell_str,
        match_UMI_median
    ) = expression_matrix(
        df,
        validated_barcodes,
        args.outdir,
        args.sample,
        id_name,
        sc_cell_barcodes,
        sc_cell_number
        )

    # downsampling
    validated_barcodes = set(validated_barcodes)
    downsample_file = args.outdir + '/' + args.sample + '_downsample.txt'
    Saturation = downsample(
        count_detail_file,
        validated_barcodes,
        downsample_file)

    # summary
    stat_file = args.outdir + '/stat.txt'
    get_summary(df, args.sample, Saturation, CB_describe, CB_total_Genes,
                CB_reads_count, reads_mapped_to_transcriptome,
                match_cell_str, match_UMI_median,
                stat_file,
                args.outdir + '/../')

    report_prepare(marked_counts_file, downsample_file, args.outdir + '/..')

    t = reporter(assay=args.assay,
                 name='count_capture_rna', sample=args.sample,
                 stat_file=args.outdir + '/stat.txt',
                 outdir=args.outdir + '/..')
    t.get_report()


def get_opts_count_capture_rna(parser, sub_program):
    if sub_program:
        s_common(parser)
        parser.add_argument("--match_dir", default=None)
        parser.add_argument("--bam", default=None)
    parser.add_argument('--genomeDir',
            help='genome directory',
            required=True)
    parser.add_argument('--cells',
        help='expected number of cells',
        default="auto")


#!/bin/env python
# coding=utf8

import os
import sys
import json
import logging
import re
import numpy as np
import pandas as pd
import glob
from scipy.io import mmwrite
from scipy.sparse import csr_matrix
from celescope.tools.utils import *
from celescope.rna.__init__ import __ASSAY__
from celescope.tools.analysisMixin import AnalysisMixin 
from celescope.tools.Step import Step


TOOLSDIR = os.path.dirname(__file__)


@add_log
def generate_matrix(gtf_file, matrix_file):

    id_name = gene_convert(gtf_file)
    matrix = pd.read_csv(matrix_file, sep="\t")

    gene_name_col = matrix.geneID.apply(lambda x: id_name[x])
    matrix.geneID = gene_name_col
    matrix = matrix.drop_duplicates(subset=["geneID"], keep="first")
    matrix = matrix.dropna()
    matrix = matrix.rename({"geneID": ""}, axis='columns')
    return matrix


@add_log
def seurat(sample, outdir, matrix_file, save_rds):
    app = TOOLSDIR + "/run_analysis.R"
    cmd = (
        f'Rscript {app} --sample {sample} --outdir {outdir} --matrix_file {matrix_file} '
        f'--save_rds {save_rds}'
    )
    seurat.logger.info(cmd)
    os.system(cmd)


@add_log
def auto_assign(sample, outdir, type_marker_tsv):
    rds = f'{outdir}/{sample}.rds'
    app = TOOLSDIR + "/auto_assign.R"
    cmd = (
        f'Rscript {app} '
        f'--rds {rds} '
        f'--type_marker_tsv {type_marker_tsv} '
        f'--outdir {outdir} '
        f'--sample {sample} '
    )
    auto_assign.logger.info(cmd)
    os.system(cmd)


class Analysis_rna(Step, AnalysisMixin):
    def __init__(self, args, step_name):
        Step.__init__(self, args, step_name)
        AnalysisMixin.__init__(self, args)
        self.matrix_file = args.matrix_file
        self.type_marker_tsv = args.type_marker_tsv
        self.auto_assign_bool = False
        self.save_rds = args.save_rds
        if args.type_marker_tsv and args.type_marker_tsv != 'None':
            self.auto_assign_bool = True
            self.save_rds = True

    def run(self):
        seurat(self.sample, self.outdir, self.matrix_file, self.save_rds)
        if self.auto_assign_bool:
            auto_assign(self.sample, self.outdir, self.type_marker_tsv)
        self.run_analysis()
        self.add_data_item(cluster_tsne=self.cluster_tsne)
        self.add_data_item(gene_tsne=self.gene_tsne)
        self.add_data_item(table_dict=self.table_dict)

        self.clean_up()


@add_log
def analysis(args):

    step_name = "analysis"
    ana = Analysis_rna(args, step_name)
    ana.run()


def get_opts_analysis(parser, sub_program):
    if sub_program:
        parser = s_common(parser)
        parser.add_argument('--matrix_file', help='matrix file', required=True)
    parser.add_argument('--save_rds', action='store_true', help='write rds to disk')
    parser.add_argument('--type_marker_tsv', help='cell type marker tsv')




import os
import sys
import json
import numpy as np
import pandas as pd
import glob
from celescope.tools.report import reporter
from celescope.tools.utils import *
from celescope.tools.Analysis import Analysis

class Analysis_tag(Analysis):

    def run(self, tsne_tag_file):
        self.cluster_tsne = self.get_cluster_tsne(colname='cluster')
        self.tsne_tag_df = pd.read_csv(tsne_tag_file, sep="\t", index_col=0)
        self.feature_tsne = self.get_cluster_tsne(colname='tag', show_tag=False, dfname='tsne_tag_df')
        self.marker_gene_table = self.marker_table()









from celescope.fusion.__init__ import __STEPS__, __ASSAY__
from celescope.tools.Multi import Multi


class Multi_fusion(Multi):

    def STAR_fusion(self, sample):
        step = 'STAR_fusion'
        cmd_line = self.get_cmd_line(step, sample)
        input_read = f'{self.outdir_dic[sample]["cutadapt"]}/{sample}_clean_2.fq{self.fq_suffix}'
        cmd = (
            f'{cmd_line} '
            f'--input_read {input_read} '
        )
        self.process_cmd(cmd, step, sample, m=self.args.starMem, x=self.args.thread)

    def count_fusion(self, sample):
        step = 'count_fusion'
        cmd_line = self.get_cmd_line(step, sample)
        bam = f'{self.outdir_dic[sample]["STAR_fusion"]}/{sample}_Aligned.sortedByCoord.out.bam'
        cmd = (
            f'{cmd_line} '
            f'--bam {bam} '
            f'--match_dir {self.col4_dict[sample]} '
        )
        self.process_cmd(cmd, step, sample, m=15, x=1)


def main():
    multi = Multi_fusion(__ASSAY__)
    multi.run()


if __name__ == '__main__':
    main()





from celescope.snp.__init__ import __STEPS__, __ASSAY__
from celescope.tools.Multi import Multi


class Multi_snp(Multi):

    def STAR(self, sample):
        step = 'STAR'
        cmd_line = self.get_cmd_line(step, sample)
        if self.args.not_consensus:
            fq = f'{self.outdir_dic[sample]["cutadapt"]}/{sample}_clean_2.fq{self.fq_suffix}'
        else:
            fq = f'{self.outdir_dic[sample]["consensus"]}/{sample}_consensus.fq'
            cmd_line += ' --consensus_fq '
   
        cmd = (
            f'{cmd_line} '
            f'--fq {fq} '
        )
        self.process_cmd(cmd, step, sample, m=self.args.starMem, x=self.args.thread)

    def target_metrics(self, sample):
        step = 'target_metrics'
        cmd_line = self.get_cmd_line(step, sample)
        bam = f'{self.outdir_dic[sample]["featureCounts"]}/{sample}_name_sorted.bam'
        cmd = (
            f'{cmd_line} '
            f'--bam {bam} '
            f'--match_dir {self.col4_dict[sample]} '
        )
        self.process_cmd(cmd, step, sample, m=2, x=1)


    def snpCalling(self, sample):
        step = 'snpCalling'
        cmd_line = self.get_cmd_line(step, sample)
        bam = f'{self.outdir_dic[sample]["featureCounts"]}/{sample}_name_sorted.bam'
        cmd = (
            f'{cmd_line} '
            f'--bam {bam} '
            f'--match_dir {self.col4_dict[sample]} '
        )
        self.process_cmd(cmd, step, sample, m=8, x=self.args.thread)

    def analysis_snp(self, sample):
        step = 'analysis_snp'
        vcf = f'{self.outdir_dic[sample]["snpCalling"]}/{sample}_merged.vcf'
        CID_file = f'{self.outdir_dic[sample]["snpCalling"]}/{sample}_CID.tsv'
        variant_count_file = f'{self.outdir_dic[sample]["snpCalling"]}/{sample}_variant_count.tsv'
        cmd_line = self.get_cmd_line(step, sample)
        cmd = (
            f'{cmd_line} '
            f'--match_dir {self.col4_dict[sample]} '
            f'--vcf {vcf} '
            f'--CID_file {CID_file} '
            f'--variant_count_file {variant_count_file} '
        )
        self.process_cmd(cmd, step, sample, m=8, x=1)


def main():
    multi = Multi_snp(__ASSAY__)
    multi.run()

if __name__ == '__main__':
    main()




import glob
import gzip
import os
import pandas as pd
import logging
import numpy as np
import sys
import argparse
import matplotlib as mpl
import re
import json
mpl.use('Agg')
from matplotlib import pyplot as plt

from celescope.tools.Step import Step, s_common
from celescope.vdj.__init__ import CHAINS
from celescope.tools.report import reporter
import celescope.tools.utils as utils
from celescope.tools.Analysis import Analysis


def report_prepare(df, outdir):

    json_file = outdir + '/.data.json'
    if not os.path.exists(json_file):
        data = {}
    else:
        fh = open(json_file)
        data = json.load(fh)
        fh.close()

    df = df.sort_values('UMI', ascending=False)
    data['CB_num'] = df[df['mark'] == 'CB'].shape[0]
    data['Cells'] = list(df.loc[df['mark'] == 'CB', 'UMI'])
    data['UB_num'] = df[df['mark'] == 'UB'].shape[0]
    data['Background'] = list(df.loc[df['mark'] == 'UB', 'UMI'])

    with open(json_file, 'w') as fh:
        json.dump(data, fh)


@utils.add_log
def count_vdj(args):

    step_name = f"{args.type}_count_vdj"
    step = Step(args, step_name)

    sample = args.sample
    match_dir = args.match_dir
    UMI_min = args.UMI_min
    outdir = args.outdir
    UMI_count_filter1_file = args.UMI_count_filter1_file
    type = args.type
    assay = args.assay
    debug = args.debug
    iUMI = int(args.iUMI)
    chains = CHAINS[type]

    # out file
    cell_confident_file = f"{outdir}/{sample}_cell_confident.tsv"
    cell_confident_count_file = f"{outdir}/{sample}_cell_confident_count.tsv"
    clonetypes_file = f"{outdir}/{sample}_clonetypes.tsv"
    match_clonetypes_file = f"{outdir}/{sample}_match_clonetypes.tsv"

    # read file
    df_UMI_count_filter1 = pd.read_csv(UMI_count_filter1_file, sep='\t')
    if (not match_dir) or (match_dir == "None"):
        match_bool = False
    else:
        match_bool = True
    if match_bool:
        match_cell_barcodes, match_cell_number = utils.read_barcode_file(match_dir)

    cell_summary_row_list = []

    # cell calling：cell calling: keep UMIs >= UMI_min
    df_UMI_sum = df_UMI_count_filter1.groupby(
        ['barcode'], as_index=False).agg({"UMI": "sum"})
    if (UMI_min == "auto"):
        rank = 20
        df_UMI_sum_sorted = df_UMI_sum.sort_values(["UMI"], ascending=False)
        rank_UMI = df_UMI_sum_sorted.iloc[rank, :]["UMI"]
        UMI_min = int(rank_UMI / 10)
    else:
        UMI_min = int(UMI_min)
    df_UMI_cell = df_UMI_sum[df_UMI_sum.UMI >= UMI_min]
    df_UMI_sum["mark"] = df_UMI_sum["UMI"].apply(
        lambda x: "CB" if (x >= UMI_min) else "UB")
    
    # data
    df = df_UMI_sum.sort_values('UMI', ascending=False)
    step.add_data_item(CB_num=df[df['mark'] == 'CB'].shape[0])
    step.add_data_item(Cells=list(df.loc[df['mark'] == 'CB', 'UMI']))
    step.add_data_item(UB_num= df[df['mark'] == 'UB'].shape[0])
    step.add_data_item(Background=list(df.loc[df['mark'] == 'UB', 'UMI']))


    cell_barcodes = set(df_UMI_cell.barcode)
    cell_number = len(cell_barcodes)
    cell_summary_row_list.append({
        "item": "Estimated Number of Cells",
        "count": cell_number,
        "total_count": cell_number,
    })

    # df_UMI_count_filter1 in cell
    df_cell = df_UMI_count_filter1[df_UMI_count_filter1.barcode.isin(
        cell_barcodes)]
    # filter2: cell wtih UMI >= iUMI of identical receptor type and CDR3
    # combinations.
    df_cell_UMI_count_filter2 = df_cell[df_cell.UMI >= iUMI]

    # cell confident
    df_cell_confident = df_cell_UMI_count_filter2[df_cell_UMI_count_filter2["chain"].isin(
        chains)]
    df_cell_confident = df_cell_confident.sort_values("UMI", ascending=False)
    df_cell_confident = df_cell_confident.groupby(
        ["barcode", "chain"], as_index=False).head(1)

    # count
    df_cell_confident_count = df_cell_confident.set_index(["barcode", "chain"])
    df_cell_confident_count = df_cell_confident_count.unstack()
    df_cell_confident_count.columns = [
        '_'.join(col) for col in df_cell_confident_count]
    df_cell_confident_count = df_cell_confident_count.reset_index()
    df_cell_confident_count.fillna(inplace=True, value="NA")

    # clonetypes
    seqs = ["aaSeqCDR3", "nSeqCDR3"]
    cols = []
    for chain in chains:
        for seq in seqs:
            cols.append("_".join([seq, chain]))

    for col in cols:
        if not (col in list(df_cell_confident_count.columns)):
            df_cell_confident_count[col] = "NA"

    df_clonetypes = df_cell_confident_count.copy()

    df_clonetypes = df_clonetypes.groupby(cols, as_index=False).agg({
        "barcode": "count"})
    # put na last
    df_clonetypes.replace('NA', np.nan, inplace=True)
    df_clonetypes.sort_values(["barcode"] + cols, ascending=False, na_position='last', inplace=True)
    df_clonetypes.replace(np.nan, 'NA', inplace=True)

    total_CDR3_barcode_number = sum(df_clonetypes.barcode)
    df_clonetypes["percent"] = df_clonetypes.barcode / \
        total_CDR3_barcode_number * 100
    df_clonetypes["percent"] = df_clonetypes["percent"].apply(
        lambda x: round(x, 2))

    # add clonetype ID
    df_clonetypes = df_clonetypes.reset_index()
    df_clonetypes["clonetype_ID"] = pd.Series(df_clonetypes.index) + 1
    df_clonetypes.drop(columns=["index"], inplace=True)

    # order
    order = ["clonetype_ID"] + cols + ["barcode", "percent"]
    df_clonetypes = df_clonetypes[order]
    df_clonetypes.rename(columns={"barcode": "barcode_count"}, inplace=True)
    # out clonetypes
    df_clonetypes.to_csv(clonetypes_file, sep="\t", index=False)

    if type == "TCR":

        UMI_col_dic = {"TRA": "UMI_TRA", "TRB": "UMI_TRB"}
        for chain in UMI_col_dic:
            UMI_col_name = UMI_col_dic[chain]
            if UMI_col_name in df_cell_confident_count.columns:
                df_cell_confident_count[UMI_col_name].replace(
                    "NA", 0, inplace=True)
                Median_chain_UMIs_per_Cell = np.median(
                    df_cell_confident_count[UMI_col_name])
            else:
                Median_chain_UMIs_per_Cell = 0
            cell_summary_row_list.append({
                "item": "Median {chain} UMIs per Cell".format(chain=chain),
                "count": Median_chain_UMIs_per_Cell,
                "total_count": np.nan
            })

        df_TRA_TRB = df_cell_confident_count[
            (df_cell_confident_count.aaSeqCDR3_TRA != "NA") &
            (df_cell_confident_count.aaSeqCDR3_TRB != "NA")
        ]
        cell_with_confident_TRA_and_TRB = df_TRA_TRB.shape[0]
        cell_summary_row_list.append({
            "item": "Cell with TRA and TRB",
            "count": cell_with_confident_TRA_and_TRB,
            "total_count": cell_number,
        })

        """
        df cell barcode filter
        intersect cell_barcodes from scRNA-Seq with barcode from TCR seq
        """
        if match_bool:
            cell_with_match_barcode = match_cell_barcodes.intersection(
                cell_barcodes)
            cell_with_match_barcode_number = len(cell_with_match_barcode)

            df_match = df_cell_confident_count[df_cell_confident_count.barcode.isin(
                match_cell_barcodes)]

            df_match_TRA_TRB = df_match[
                (df_match.aaSeqCDR3_TRA != "NA") &
                (df_match.aaSeqCDR3_TRB != "NA")
            ]
            match_cell_with_TRA_and_TRB = df_match_TRA_TRB.shape[0]

            cell_summary_row_list.append({
                "item": "Cell with Barcode Match",
                "count": cell_with_match_barcode_number,
                "total_count": cell_number,
            })
            cell_summary_row_list.append({
                "item": "Cell with Barcode Match, TRA and TRB",
                "count": match_cell_with_TRA_and_TRB,
                "total_count": cell_number,
            })

    # BCR
    elif type == "BCR":

        UMI_col_dic = {"IGH": "UMI_IGH", "IGL": "UMI_IGL", "IGK": "UMI_IGK"}
        for chain in UMI_col_dic:
            UMI_col_name = UMI_col_dic[chain]
            if UMI_col_name in df_cell_confident_count.columns:
                df_cell_confident_count[UMI_col_name].replace(
                    "NA", 0, inplace=True)
                df_cell_confident_count_over_zero = df_cell_confident_count[
                    df_cell_confident_count[UMI_col_name] > 0
                ]
                Median_chain_UMIs_per_Cell = np.median(
                    df_cell_confident_count_over_zero[UMI_col_name])
            else:
                Median_chain_UMIs_per_Cell = 0
            cell_summary_row_list.append({
                "item": "Median {chain} UMIs per Cell".format(chain=chain),
                "count": Median_chain_UMIs_per_Cell,
                "total_count": np.nan})

        df_heavy_and_light = df_cell_confident_count[
            (df_cell_confident_count.aaSeqCDR3_IGH != "NA") &
            (
                (df_cell_confident_count.aaSeqCDR3_IGL != "NA") |
                (df_cell_confident_count.aaSeqCDR3_IGK != "NA")
            )
        ]
        Cell_with_Heavy_and_Light_Chain = df_heavy_and_light.shape[0]
        cell_summary_row_list.append({
            "item": "Cell with Heavy and Light Chain",
            "count": Cell_with_Heavy_and_Light_Chain,
            "total_count": cell_number
        })

        """
        df cell barcode filter
        intersect cell_barcodes from normal scRNA-Seq with barcode from BCR seq
        """
        if match_bool:
            cell_with_match_barcode = match_cell_barcodes.intersection(
                cell_barcodes)
            cell_with_match_barcode_number = len(cell_with_match_barcode)

            df_match = df_cell_confident_count[df_cell_confident_count.barcode.isin(
                match_cell_barcodes)]

            # median match UMI
            df_match_heavy_light = df_match[
                (df_match.aaSeqCDR3_IGH != "NA") &
                (
                    (df_match.aaSeqCDR3_IGL != "NA") |
                    (df_match.aaSeqCDR3_IGK != "NA")
                )
            ]
            match_cell_with_heavy_and_light = df_match_heavy_light.shape[0]

            cell_summary_row_list.append({
                "item": "Cell with Barcode Match ",
                "count": cell_with_match_barcode_number,
                "total_count": cell_number
            })
            cell_summary_row_list.append({
                "item": "Cell with Barcode Match, Heavy and Light Chain",
                "count": match_cell_with_heavy_and_light,
                "total_count": cell_number
            })

    if match_bool:
        """
        df_match_clonetypes
        """
        df_match_clonetypes = df_match.groupby(cols, as_index=False).agg({
            "barcode": "count"})
        total_match_CDR3_barcode_number = sum(
            df_match_clonetypes.barcode)
        df_match_clonetypes["percent"] = df_match_clonetypes.barcode / \
            total_match_CDR3_barcode_number * 100
        df_match_clonetypes["percent"] = df_match_clonetypes["percent"].apply(
            lambda x: round(x, 2)
        )
        df_match_clonetypes.rename(columns={"barcode": "barcode_count"}, inplace=True)
        df_match_clonetypes = df_match_clonetypes.merge(
            df_clonetypes, on=cols, how='left', suffixes=('', '_y'))
        # order and drop duplicated cols
        order = ["clonetype_ID"] + cols + ["barcode_count", "percent"]
        df_match_clonetypes = df_match_clonetypes[order]
        df_match_clonetypes.sort_values(["barcode_count", "clonetype_ID"], ascending=[False,True], inplace=True)
        df_match_clonetypes.to_csv(
            match_clonetypes_file, sep="\t", index=False)

    df_mergeID = pd.merge(df_cell_confident_count,
                          df_clonetypes, how="left", on=cols)
    df_mergeID.sort_values(["clonetype_ID", "barcode"], inplace=True)
    # output df_cell_confident_count
    df_mergeID.to_csv(cell_confident_count_file, sep="\t", index=False)
    df_mergeID = df_mergeID[["barcode", "clonetype_ID"]]
    df_cell_confident_with_ID = pd.merge(
        df_cell_confident, df_mergeID, how="left", on="barcode")
    df_cell_confident_with_ID.sort_values(
        ["clonetype_ID", "barcode", "chain"], inplace=True)
    # output df_cell_confident
    df_cell_confident_with_ID.to_csv(
        cell_confident_file, sep="\t", index=False)

    # summary file
    cell_summary = pd.DataFrame(cell_summary_row_list, columns=[
                                "item", "count", "total_count"])
    cell_summary["count"] = cell_summary["count"].apply(int)
    cell_summary["percent"] = cell_summary["count"] / \
        (cell_summary.total_count.astype("float")) * 100
    cell_summary["percent"] = cell_summary["percent"].apply(
        lambda x: round(x, 2))
    cell_summary["count"] = cell_summary["count"].apply(utils.format_number)

    def percent_str_func(row):
        need_percent = bool(
            re.search("Cell with", row["item"], flags=re.IGNORECASE))
        if need_percent:
            return "(" + str(row["percent"]) + "%)"
        else:
            return ""
    cell_summary["percent_str"] = cell_summary.apply(
        lambda row: percent_str_func(row), axis=1)

    # stat file
    def gen_stat(summary, stat_file):
        stat = summary
        stat["new_count"] = stat["count"].astype(str) + stat["percent_str"]
        stat = stat.loc[:, ["item", "new_count"]]
        stat.to_csv(stat_file, sep=":", header=None, index=False)

    cell_stat_file = "{}/stat.txt".format(outdir)
    gen_stat(cell_summary, cell_stat_file)
    step.add_data_item(iUMI=iUMI)

    # cloneytpes table
    def format_table(df_clonetypes):
        df_table = df_clonetypes.copy()
        df_table["percent"] = df_table["percent"].apply(
            lambda x: str(x) + "%")
        seqs = ["aaSeqCDR3"]
        cols = []
        for chain in chains:
            for seq in seqs:
                cols.append("_".join([seq, chain]))
        df_table_cols = ["clonetype_ID"] + cols + ["barcode_count", "percent"]
        df_table = df_table[df_table_cols]
        table_header = ["Clonetype_ID"] + cols + ["Frequency", "Percent"]
        return df_table, table_header

    df_table, table_header = format_table(df_clonetypes)
    title = 'Clonetypes'
    if match_bool:
        df_table, table_header = format_table(df_match_clonetypes)
        title = 'Match Clonetypes'

    table_dict = step.get_table(title, 'clonetypes_table', df_table)
    step.add_data_item(table_dict=table_dict)
    
    # other_metrics_file
    """
    if len(other_metrics_row_list) != 0:
        other_metrics = pd.DataFrame(other_metrics_row_list,columns=["item","count"])
        other_metrics.to_csv(other_metrics_file,sep=":",header=None,index=False)
    """

    step.clean_up()


def get_opts_count_vdj(parser, sub_program):
    if sub_program:
        parser = s_common(parser)
        parser.add_argument("--UMI_count_filter1_file", required=True)
        parser.add_argument("--match_dir", default=None)
    parser.add_argument("--type", required=True)
    parser.add_argument('--UMI_min', dest='UMI_min',
                        help='minimum UMI number to filter', default="auto")
    parser.add_argument(
        '--iUMI', help='minimum number of UMI of identical receptor type and CDR3', default=1)
